#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#

from pysmi.codegen.base import AbstractCodeGen


def capfirst(text):
    if not text:
        return text

    return text[0].upper() + text[1:]


def bitstring(bits):
    mask = sum(1 << bit for bit in bits)

    # The left-most character of the returned string is for bit number zero,
    # so reverse the bits, while also stripping off the "0b" prefix.
    return bin(mask)[:1:-1]


def pythonsym(symbol):
    return AbstractCodeGen.trans_opers(symbol)


def pythonstr(text):
    if "\n" in text or "\r" in text:
        return '"""\\\n' + text.replace("\\", "\\\\") + '"""'
    else:
        return '"' + text.replace("\\", "\\\\") + '"'



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#

#
# Preconfigured sets of parser options.
# Individual options could be used in certain combinations.
#
import warnings


smi_v2 = {}  # TODO: move strict mode here.

smi_v1 = smi_v2.copy()
smi_v1.update(supportSmiV1Keywords=True, supportIndex=True)

smi_v1_relaxed = smi_v1.copy()
smi_v1_relaxed.update(
    commaAtTheEndOfImport=True,
    commaAtTheEndOfSequence=True,
    mixOfCommasAndSpaces=True,
    uppercaseIdentifier=True,
    lowcaseIdentifier=True,
    curlyBracesAroundEnterpriseInTrap=True,
    noCells=True,
)

# Compatibility API
deprecated_attributes = {
    "smiV1Relaxed": "smi_v1_relaxed",
    "smiV1": "smi_v1",
    "smiV2": "smi_v2",
}


def __getattr__(attr: str):
    if new_attr := deprecated_attributes.get(attr):
        warnings.warn(
            f"{attr} is deprecated. Please use {new_attr} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return globals()[new_attr]
    raise AttributeError(f"module '{__name__}' has no attribute '{attr}'")



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import os
import sys
from collections import OrderedDict

import jinja2
from pysmi import debug, error
from pysmi.codegen import jfilters
from pysmi.codegen.intermediate import IntermediateCodeGen
from pysmi.mibinfo import MibInfo


class PySnmpCodeGen(IntermediateCodeGen):
    """Turns MIB AST into pysnmp Python code.

    Builds Python module representing MIB module in terms of pysnmp objects
    from MIB supplied in form of an Abstract Syntax Tree on input.

    Instance of this class is supposed to be passed to *MibCompiler*,
    the rest is internal to *MibCompiler*.
    """

    defaultMibPackages = ("pysnmp.smi.mibs", "pysnmp_mibs")

    TEMPLATE_NAME = "pysnmp/mib-definitions.j2"

    SMI_OBJECTS = {
        "MODULE-IDENTITY": ["ModuleIdentity"],
        "OBJECT-TYPE": ["MibScalar", "MibTable", "MibTableRow", "MibTableColumn"],
        "NOTIFICATION-TYPE": ["NotificationType"],
        "TEXTUAL-CONVENTION": ["TextualConvention"],
        "MODULE-COMPLIANCE": ["ModuleCompliance"],
        "OBJECT-GROUP": ["ObjectGroup"],
        "NOTIFICATION-GROUP": ["NotificationGroup"],
        "AGENT-CAPABILITIES": ["AgentCapabilities"],
        "OBJECT-IDENTITY": ["ObjectIdentity"],
        "TRAP-TYPE": ["NotificationType"],
        "BITS": ["Bits"],
    }

    SMI_TYPES = {
        "COUNTER32": "Counter32",
        "COUNTER64": "Counter64",
        "GAUGE32": "Gauge32",
        "INTEGER": "Integer32",  # XXX
        "INTEGER32": "Integer32",
        "IPADDRESS": "IpAddress",
        "NETWORKADDRESS": "IpAddress",
        "OBJECT IDENTIFIER": "ObjectIdentifier",
        "OCTET STRING": "OctetString",
        "OPAQUE": "Opaque",
        "TIMETICKS": "TimeTicks",
        "UNSIGNED32": "Unsigned32",
        "Counter": "Counter32",
        "Gauge": "Gauge32",
        "NetworkAddress": "IpAddress",  # RFC1065-SMI, RFC1155-SMI -> SNMPv2-SMI
        "nullSpecific": "zeroDotZero",  # RFC1158-MIB -> SNMPv2-SMI
        "ipRoutingTable": "ipRouteTable",  # RFC1158-MIB -> RFC1213-MIB
        "snmpEnableAuthTraps": "snmpEnableAuthenTraps",  # RFC1158-MIB -> SNMPv2-MIB
    }

    # never compile these, they either:
    # - define MACROs (implementation supplies them)
    # - or carry conflicting OIDs (so that all IMPORT's of them will be rewritten)
    # - or have manual fixes
    # - or import base ASN.1 types from implementation-specific MIBs
    fakeMibs = ("ASN1", "ASN1-ENUMERATION", "ASN1-REFINEMENT")
    baseMibs = (
        "PYSNMP-USM-MIB",
        "SNMP-FRAMEWORK-MIB",
        "SNMP-TARGET-MIB",
        "TRANSPORT-ADDRESS-MIB",
        "INET-ADDRESS-MIB",
    ) + IntermediateCodeGen.baseMibs

    def gen_code(self, ast, symbolTable, **kwargs) -> "tuple[MibInfo, str]":
        mibInfo, context = IntermediateCodeGen.gen_code(
            self, ast, symbolTable, **kwargs
        )

        # Adapt intermediate context to pysnmp template requirements

        # Translate SMI objects names in IMPORT

        imports = OrderedDict()

        for module, symbols in context.get("imports", {}).items():
            if not isinstance(symbols, list):
                continue

            imports[module] = []

            for symbol in symbols:
                if symbol in self.SMI_OBJECTS:
                    imports[module].extend(self.SMI_OBJECTS[symbol])
                else:
                    imports[module].append(symbol)

        context["imports"] = imports

        # Turn string OIDs into tuples which is native to pysnmp
        # TODO: we should make intermediate format producing tuples

        def translate_oids(dct):
            for key, value in tuple(dct.items()):
                if isinstance(value, dict):
                    translate_oids(value)
                elif key == "oid":
                    dct[key] = tuple(int(x) for x in value.split("."))

        translate_oids(context)

        # Translate SMI types into pysnmp class names

        # Sort Managed Objects by OID
        objects = OrderedDict()

        for symbol, definition in sorted(
            context.items(), key=lambda x: x[1].get("oid", ())
        ):
            objects[symbol] = definition

        context = objects

        # Render Python code

        searchPath = [os.path.join(os.path.dirname(__file__), "templates")]

        # TODO: add unit test on custom template

        dstTemplate = kwargs.get("dstTemplate")
        if dstTemplate:
            searchPath.insert(0, os.path.dirname(os.path.abspath(dstTemplate)))

        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(searchPath),
            trim_blocks=True,
            lstrip_blocks=True,
        )

        env.filters["capfirst"] = jfilters.capfirst
        env.filters["bitstring"] = jfilters.bitstring
        env.filters["pythonsym"] = jfilters.pythonsym
        env.filters["pythonstr"] = jfilters.pythonstr

        try:
            tmpl = env.get_template(dstTemplate or self.TEMPLATE_NAME)
            text = tmpl.render(mib=context)

        except jinja2.exceptions.TemplateError:
            err = sys.exc_info()[1]
            raise error.PySmiCodegenError(f"Jinja template rendering error: {err}")

        debug.logger & debug.FLAG_CODEGEN and debug.logger(
            f"canonical MIB name {mibInfo.name} ({mibInfo.identity}), imported MIB(s) {','.join(mibInfo.imported) or '<none>'}, rendered from {dstTemplate}, Python code size {len(text)} bytes"
        )

        return mibInfo, text


# backward compatibility
baseMibs = PySnmpCodeGen.baseMibs  # noqa: N816
fakeMibs = PySnmpCodeGen.fakeMibs  # noqa: N816



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import importlib.machinery
import importlib.util
import os
import struct
import time

from pysmi import debug, error
from pysmi.compat import decode
from pysmi.searcher.base import AbstractSearcher
from pysmi.searcher.pyfile import PyFileSearcher

PY_MAGIC_NUMBER = importlib.util.MAGIC_NUMBER
SOURCE_SUFFIXES = importlib.machinery.SOURCE_SUFFIXES
BYTECODE_SUFFIXES = importlib.machinery.BYTECODE_SUFFIXES


class PyPackageSearcher(AbstractSearcher):
    """Figures out if given Python module (source or bytecode) exists in given Python package.

    Python package must be importable.
    """

    def __init__(self, package):
        """Create an instance of *PyPackageSearcher* bound to specific Python package.

        Args:
            package (str): name of the Python package to look up Python
                           modules at.
        """
        self._package = package
        self.__loader = None

    def __str__(self):
        """Return a string representation of the instance."""
        return f'{self.__class__.__name__}{{"{self._package}"}}'

    @staticmethod
    def _parse_dos_time(dosdate, dostime):
        t = (
            ((dosdate >> 9) & 0x7F) + 1980,  # year
            ((dosdate >> 5) & 0x0F),  # month
            dosdate & 0x1F,  # mday
            (dostime >> 11) & 0x1F,  # hour
            (dostime >> 5) & 0x3F,  # min
            (dostime & 0x1F) * 2,  # sec
            -1,  # wday
            -1,  # yday
            -1,
        )  # dst
        return time.mktime(t)

    def file_exists(self, mibname, mtime, rebuild=False):
        if rebuild:
            debug.logger & debug.FLAG_SEARCHER and debug.logger(
                f"pretend {mibname} is very old"
            )
            return

        mibname = decode(mibname)

        try:
            p = __import__(self._package, globals(), locals(), ["__init__"])

            if hasattr(p, "__loader__") and hasattr(p.__loader__, "_files"):
                self.__loader = p.__loader__
                self._package = self._package.replace(".", os.sep)
                debug.logger & debug.FLAG_SEARCHER and debug.logger(
                    f"{self._package} is an importable egg at {os.path.split(p.__file__)[0]}"
                )

            elif hasattr(p, "__file__"):
                debug.logger & debug.FLAG_SEARCHER and debug.logger(
                    f"{self._package} is not an egg, trying it as a package directory"
                )
                return PyFileSearcher(os.path.split(p.__file__)[0]).file_exists(
                    mibname, mtime, rebuild=rebuild
                )

            else:
                raise error.PySmiFileNotFoundError(
                    f"{self._package} is neither importable nor a file", searcher=self
                )

        except ImportError:
            raise error.PySmiFileNotFoundError(
                f"{self._package} is not importable, trying as a path", searcher=self
            )

        for pySfx in BYTECODE_SUFFIXES:
            f = os.path.join(self._package, mibname.upper()) + pySfx

            if f not in self.__loader._files:
                debug.logger & debug.FLAG_SEARCHER and debug.logger(
                    f"{f} is not in {self._package}"
                )
                continue

            pyData = self.__loader.get_data(f)
            if pyData[:4] == PY_MAGIC_NUMBER:
                pyData = pyData[4:]
                pyTime = struct.unpack("<L", pyData[:4])[0]
                debug.logger & debug.FLAG_SEARCHER and debug.logger(
                    f"found {f}, mtime {time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(pyTime))}"
                )
                if pyTime >= mtime:
                    raise error.PySmiFileNotModifiedError()
                else:
                    raise error.PySmiFileNotFoundError(
                        f"older file {mibname} exists", searcher=self
                    )

            else:
                debug.logger & debug.FLAG_SEARCHER and debug.logger(f"bad magic in {f}")
                continue

        for pySfx in SOURCE_SUFFIXES:
            f = os.path.join(self._package, mibname.upper()) + pySfx

            if f not in self.__loader._files:
                debug.logger & debug.FLAG_SEARCHER and debug.logger(
                    f"{f} is not in {self._package}"
                )
                continue

            pyTime = self._parse_dos_time(
                self.__loader._files[f][6], self.__loader._files[f][5]
            )

            debug.logger & debug.FLAG_SEARCHER and debug.logger(
                f"found {f}, mtime {time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(pyTime))}"
            )
            if pyTime >= mtime:
                raise error.PySmiFileNotModifiedError()
            else:
                raise error.PySmiFileNotFoundError(
                    f"older file {mibname} exists", searcher=self
                )

        raise error.PySmiFileNotFoundError(f"no file {mibname} found", searcher=self)



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import getpass
import platform
import sys
import time
import warnings

from pysmi import __name__ as package_name
from pysmi import __version__ as package_version
from pysmi import debug
from pysmi import error
from pysmi.borrower.base import AbstractBorrower
from pysmi.codegen.base import AbstractCodeGen
from pysmi.codegen.symtable import SymtableCodeGen
from pysmi.mibinfo import MibInfo
from pysmi.reader.base import AbstractReader
from pysmi.searcher.base import AbstractSearcher
from pysmi.writer.base import AbstractWriter


class MibStatus(str):
    """Indicate MIB transformation result.

    *MibStatus* is a subclass of Python string type. Some additional
    attributes may be set to indicate the details.

    The following *MibStatus* class instances are defined:

    * *compiled* - MIB is successfully transformed
    * *untouched* - fresh transformed version of this MIB already exisits
    * *failed* - MIB transformation failed. *error* attribute carries details.
    * *unprocessed* - MIB transformation required but waived for some reason
    * *missing* - ASN.1 MIB source can't be found
    * *borrowed* - MIB transformation failed but pre-transformed version was used
    """

    def set_options(self, **kwargs):
        n = self.__class__(self)
        for k in kwargs:
            setattr(n, k, kwargs[k])
        return n


status_compiled = MibStatus("compiled")
status_untouched = MibStatus("untouched")
status_failed = MibStatus("failed")
status_unprocessed = MibStatus("unprocessed")
status_missing = MibStatus("missing")
status_borrowed = MibStatus("borrowed")


class MibCompiler:
    """Top-level, user-facing, composite MIB compiler object.

    MibCompiler implements high-level MIB transformation processing logic.
    It executes its actions by calling the following specialized objects:

      * *readers* - to acquire ASN.1 MIB data
      * *searchers* - to see if transformed MIB already exists and no processing is necessary
      * *parser* - to parse ASN.1 MIB into AST
      * *code generator* - to perform actual MIB transformation
      * *borrowers* - to fetch pre-transformed MIB if transformation is impossible
      * *writer* - to store transformed MIB data

    Required components must be passed to MibCompiler on instantiation. Those
    components are: *parser*, *codegenerator* and *writer*.

    Optional components could be set or modified at later phases of MibCompiler
    life. Unlike singular, required components, optional one can be present
    in sequences to address many possible sources of data. They are
    *readers*, *searchers* and *borrowers*.
    """

    indexFile = "index"
    _searchers: list[AbstractSearcher]
    _sources: list[AbstractReader]
    _borrowers: list[AbstractBorrower]
    _parsedMibs: dict[str, tuple]

    failedMibs: dict[str, error.PySmiError]

    def __init__(self, parser, codegen: AbstractCodeGen, writer: AbstractWriter):
        """Creates an instance of *MibCompiler* class.

        Args:
            parser: ASN.1 MIB parser object
            codegen: MIB transformation object
            writer: transformed MIB storing object
        """
        self._parser = parser
        self._codegen = codegen
        self._symbolgen = SymtableCodeGen()
        self._writer = writer
        self._sources = []
        self._searchers = []
        self._borrowers = []

    def add_sources(self, *sources):
        """Add more ASN.1 MIB source repositories.

        MibCompiler.compile will invoke each of configured source objects
        in order of their addition asking each to fetch MIB module specified
        by name.

        Args:
            sources: reader object(s)

        Returns:
            reference to itself (can be used for call chaining)

        """
        self._sources.extend(sources)

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"current MIB source(s): {', '.join(map(str, self._sources))}"
        )

        return self

    def add_searchers(self, *searchers):
        """Add more transformed MIBs repositories.

        MibCompiler.compile will invoke each of configured searcher objects
        in order of their addition asking each if already transformed MIB
        module already exists and is more recent than specified.

        Args:
            searchers: searcher object(s)

        Returns:
            reference to itself (can be used for call chaining)

        """
        self._searchers.extend(searchers)

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"current compiled MIBs location(s): {', '.join(map(str, self._searchers))}"
        )

        return self

    def add_borrowers(self, *borrowers):
        """Add more transformed MIBs repositories to borrow MIBs from.

        Whenever MibCompiler.compile encounters MIB module which neither of
        the *searchers* can find or fetched ASN.1 MIB module can not be
        parsed (due to syntax errors), these *borrowers* objects will be
        invoked in order of their addition asking each if already transformed
        MIB can be fetched (borrowed).

        Args:
            borrowers: borrower object(s)

        Returns:
            reference to itself (can be used for call chaining)

        """
        self._borrowers.extend(borrowers)

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"current MIB borrower(s): {', '.join(map(str, self._borrowers))}"
        )

        return self

    def _get_system_info(self):
        # Gather platform information
        platform_info = (
            platform.system(),  # Operating system (e.g., 'Linux', 'Windows', 'Darwin')
            platform.node(),  # Hostname
            platform.release(),  # Version of the OS
        )

        # Gather user information
        user_info = (getpass.getuser(),)  # Current logged-in user

        return platform_info, user_info

    def compile(self, *mibnames, **options):
        """Transform requested and possibly referred MIBs.

        The *compile* method should be invoked when *MibCompiler* object
        is operational meaning at least *sources* are specified.

        Once called with a MIB module name, *compile* will:

        * fetch ASN.1 MIB module with given name by calling *sources*
        * make sure no such transformed MIB already exists (with *searchers*)
        * parse ASN.1 MIB text with *parser*
        * perform actual MIB transformation into target format with *code generator*
        * may attempt to borrow pre-transformed MIB through *borrowers*
        * write transformed MIB through *writer*

        The above sequence will be performed for each MIB name given in
        *mibnames* and may be performed for all MIBs referred to from
        MIBs being processed.

        Args:
            mibnames: list of ASN.1 MIBs names
            options: options that affect the way PySMI components work

        Returns:
            A dictionary of MIB module names processed (keys) and *MibStatus*
            class instances (values)

        """
        processed = {}
        parsedMibs = {}
        failedMibs = {}
        borrowedMibs = {}
        builtMibs = {}
        symbolTableMap = {}
        mibsToParse = [x for x in mibnames]
        canonicalMibNames = {}
        seenMibNames = set()

        while mibsToParse:
            mibname = mibsToParse.pop(0)

            if mibname in parsedMibs:
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"MIB {mibname} already parsed"
                )
                continue

            if mibname in failedMibs:
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"MIB {mibname} already failed"
                )
                continue

            if mibname in seenMibNames:
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"MIB {mibname} already seen (cyclic dependency)"
                )
                continue

            seenMibNames.add(mibname)

            for source in self._sources:
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"trying source {source}"
                )

                try:
                    fileInfo, fileData = source.get_data(mibname)

                    for mibTree in self._parser.parse(fileData):
                        mibInfo, symbolTable = self._symbolgen.gen_code(
                            mibTree, symbolTableMap
                        )

                        symbolTableMap[mibInfo.name] = symbolTable

                        parsedMibs[mibInfo.name] = fileInfo, mibInfo, mibTree

                        if mibname in failedMibs:
                            del failedMibs[mibname]

                        mibsToParse.extend(mibInfo.imported)

                        if fileInfo.name in mibnames:
                            if mibInfo.name not in canonicalMibNames:
                                canonicalMibNames[mibInfo.name] = []
                            canonicalMibNames[mibInfo.name].append(fileInfo.name)

                        debug.logger & debug.FLAG_COMPILER and debug.logger(
                            f"{mibInfo.name} ({mibname}) read from {fileInfo.path}, immediate dependencies: {', '.join(mibInfo.imported) or '<none>'}"
                        )

                    break
                except UnicodeDecodeError:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"http exception {mibname} found at {source}"
                    )
                    continue

                except error.PySmiReaderFileNotFoundError:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"no {mibname} found at {source}"
                    )
                    continue

                except error.PySmiError as exc:
                    exc.source = source
                    exc.mibname = mibname
                    exc.msg += f" at MIB {mibname}"

                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"{options.get('ignoreErrors') and 'ignoring ' or 'failing on '} {exc} from {source}"
                    )

                    failedMibs[mibname] = exc

                    processed[mibname] = status_failed.set_options(error=exc)

            else:
                exc = error.PySmiError(f"MIB source {mibname} not found")
                exc.mibname = mibname
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"no {mibname} found anywhere"
                )

                if mibname not in failedMibs:
                    failedMibs[mibname] = exc

                if mibname not in processed:
                    processed[mibname] = status_missing

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"MIBs analyzed {len(parsedMibs)}, MIBs failed {len(failedMibs)}"
        )

        #
        # See what MIBs need generating
        #

        for mibname in tuple(parsedMibs):
            fileInfo, mibInfo, mibTree = parsedMibs[mibname]

            debug.logger & debug.FLAG_COMPILER and debug.logger(
                f"checking if {mibname} requires updating"
            )

            for searcher in self._searchers:
                try:
                    searcher.file_exists(
                        mibname, fileInfo.mtime, rebuild=options.get("rebuild")  # type: ignore
                    )

                except error.PySmiFileNotFoundError:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"no compiled MIB {mibname} available through {searcher}"
                    )
                    continue

                except error.PySmiFileNotModifiedError:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"will be using existing compiled MIB {mibname} found by {searcher}"
                    )
                    del parsedMibs[mibname]
                    processed[mibname] = status_untouched
                    break

                except error.PySmiError as exc:
                    exc.searcher = searcher
                    exc.mibname = mibname
                    exc.msg += f" at MIB {mibname}"
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"error from {searcher}: {exc}"
                    )
                    continue

            else:
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"no suitable compiled MIB {mibname} found anywhere"
                )

                if options.get("noDeps") and mibname not in canonicalMibNames:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"excluding imported MIB {mibname} from code generation"
                    )
                    del parsedMibs[mibname]
                    processed[mibname] = status_untouched
                    continue

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"MIBs parsed {len(parsedMibs)}, MIBs failed {len(failedMibs)}"
        )

        #
        # Generate code for parsed MIBs
        #

        for mibname in parsedMibs.copy():
            fileInfo, mibInfo, mibTree = parsedMibs[mibname]

            debug.logger & debug.FLAG_COMPILER and debug.logger(
                f"compiling {mibname} read from {fileInfo.path}"
            )

            platform_info, user_info = self._get_system_info()

            comments = [
                f"ASN.1 source {fileInfo.path}",
                f"Produced by {package_name}-{package_version} at {time.asctime()}",
                f"On host {platform_info[1]} platform {platform_info[0]} version {platform_info[2]} by user {user_info[0]}",
                f"Using Python version {sys.version.splitlines()[0]}",
            ]

            try:
                mibInfo, mibData = self._codegen.gen_code(
                    mibTree,
                    symbolTableMap,
                    comments=comments,
                    dstTemplate=options.get("dstTemplate"),
                    genTexts=options.get("genTexts"),
                    textFilter=options.get("textFilter"),
                )

                builtMibs[mibname] = fileInfo, mibInfo, mibData
                del parsedMibs[mibname]

                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"{mibname} read from {fileInfo.path} and compiled by {self._writer}"
                )

            except error.PySmiError as exc:
                exc.handler = self._codegen
                exc.mibname = mibname
                exc.msg += f" at MIB {mibname}"

                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"error from {self._codegen}: {exc}"
                )

                processed[mibname] = status_failed.set_options(error=exc)

                failedMibs[mibname] = exc
                del parsedMibs[mibname]

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"MIBs built {len(parsedMibs)}, MIBs failed {len(failedMibs)}"
        )

        #
        # Try to borrow pre-compiled MIBs for failed ones
        #

        for mibname in failedMibs.copy():
            if options.get("noDeps") and mibname not in canonicalMibNames:
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"excluding imported MIB {mibname} from borrowing"
                )
                continue

            for borrower in self._borrowers:
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"trying to borrow {mibname} from {borrower}"
                )
                try:
                    fileInfo, fileData = borrower.get_data(
                        mibname, genTexts=options.get("genTexts")
                    )

                    borrowedMibs[mibname] = (
                        fileInfo,
                        MibInfo(name=mibname, imported=[]),
                        fileData,
                    )

                    del failedMibs[mibname]

                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"{mibname} borrowed with {borrower}"
                    )
                    break

                except error.PySmiError:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"error from {borrower}: {sys.exc_info()[1]}"
                    )

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"MIBs available for borrowing {len(borrowedMibs)}, MIBs failed {len(failedMibs)}"
        )

        #
        # See what MIBs need borrowing
        #

        for mibname in borrowedMibs.copy():
            debug.logger & debug.FLAG_COMPILER and debug.logger(
                f"checking if failed MIB {mibname} requires borrowing"
            )

            fileInfo, mibInfo, mibData = borrowedMibs[mibname]

            for searcher in self._searchers:
                try:
                    searcher.file_exists(
                        mibname, fileInfo.mtime, rebuild=options.get("rebuild")  # type: ignore
                    )

                except error.PySmiFileNotFoundError:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"no compiled MIB {mibname} available through {searcher}"
                    )
                    continue

                except error.PySmiFileNotModifiedError:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"will be using existing compiled MIB {mibname} found by {searcher}"
                    )
                    del borrowedMibs[mibname]
                    processed[mibname] = status_untouched
                    break

                except error.PySmiError as exc:
                    exc.searcher = searcher
                    exc.mibname = mibname
                    exc.msg += f" at MIB {mibname}"

                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"error from {searcher}: {exc}"
                    )

                    continue
            else:
                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"no suitable compiled MIB {mibname} found anywhere"
                )

                if options.get("noDeps") and mibname not in canonicalMibNames:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"excluding imported MIB {mibname} from borrowing"
                    )
                    processed[mibname] = status_untouched

                else:
                    debug.logger & debug.FLAG_COMPILER and debug.logger(
                        f"will borrow MIB {mibname}"
                    )
                    builtMibs[mibname] = borrowedMibs[mibname]

                    processed[mibname] = status_borrowed.set_options(
                        path=fileInfo.path, file=fileInfo.file, alias=fileInfo.name
                    )

                del borrowedMibs[mibname]

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"MIBs built {len(builtMibs)}, MIBs failed {len(failedMibs)}"
        )

        #
        # We could attempt to ignore missing/failed MIBs
        #

        if failedMibs and not options.get("ignoreErrors"):
            debug.logger & debug.FLAG_COMPILER and debug.logger(
                "failing with problem MIBs: " + ", ".join(failedMibs)
            )

            for mibname in builtMibs:
                processed[mibname] = status_unprocessed

            return processed

        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"proceeding with built MIBs {', '.join(builtMibs)}, failed MIBs {', '.join(failedMibs)}"
        )

        #
        # Store compiled MIBs
        #

        for mibname in builtMibs.copy():
            fileInfo, mibInfo, mibData = builtMibs[mibname]

            try:
                if options.get("writeMibs", True):
                    self._writer.put_data(
                        mibname, mibData, dryRun=options.get("dryRun")  # type: ignore
                    )

                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"{mibname} stored by {self._writer}"
                )

                del builtMibs[mibname]

                if mibname not in processed:
                    processed[mibname] = status_compiled.set_options(
                        path=fileInfo.path,
                        file=fileInfo.file,
                        alias=fileInfo.name,
                        oid=mibInfo.oid,
                        oids=mibInfo.oids,
                        identity=mibInfo.identity,
                        revision=mibInfo.revision,
                        enterprise=mibInfo.enterprise,
                        compliance=mibInfo.compliance,
                    )

            except error.PySmiError as exc:
                exc.handler = self._codegen
                exc.mibname = mibname
                exc.msg += f" at MIB {mibname}"

                debug.logger & debug.FLAG_COMPILER and debug.logger(
                    f"error {exc} from {self._writer}"
                )

                processed[mibname] = status_failed.set_options(error=exc)
                failedMibs[mibname] = exc
                del builtMibs[mibname]

        modified_mibs = [
            x for x in processed if processed[x] in ("compiled", "borrowed")
        ]
        debug.logger & debug.FLAG_COMPILER and debug.logger(
            f"MIBs modified: {', '.join(modified_mibs)}"
        )

        return processed

    def build_index(self, processedMibs, **options):
        platform_info, user_info = self._get_system_info()

        comments = [
            f"Produced by {package_name}-{package_version} at {time.asctime()}",
            f"On host {platform_info[1]} platform {platform_info[0]} version {platform_info[2]} by user {user_info[0]}",
            f"Using Python version {sys.version.splitlines()[0]}",
        ]

        try:
            self._writer.put_data(
                self.indexFile,
                self._codegen.genIndex(
                    processedMibs,
                    comments=comments,
                    old_index_data=self._writer.get_data(self.indexFile),
                ),
                dryRun=options.get("dryRun"),  # type: ignore
            )
        except error.PySmiError as exc:
            exc.msg += f" at MIB index {self.indexFile}"

            debug.logger & debug.FLAG_COMPILER and debug.logger(
                f"error {exc} when building {self.indexFile}"
            )

            if options.get("ignoreErrors"):
                return

            raise exc

    # compatibility with legacy code
    # Old to new attribute mapping
    deprecated_attributes = {
        "addSources": "add_sources",
        "addSearchers": "add_searchers",
        "addBorrowers": "add_borrowers",
    }

    def __getattr__(self, attr: str):
        """Handle deprecated attributes."""
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )



"""
Compile MIBs into JSON
++++++++++++++++++++++

Look up specific ASN.1 MIBs at configured Web sites,
compile them into JSON documents and print them out to stdout.

Try to support both SMIv1 and SMIv2 flavors of SMI as well as
popular deviations from official syntax found in the wild.
"""  #
from pysmi.reader import FileReader, HttpReader
from pysmi.searcher import StubSearcher
from pysmi.writer import CallbackWriter
from pysmi.parser import SmiStarParser
from pysmi.codegen import JsonCodeGen
from pysmi.compiler import MibCompiler
import os

# from pysmi import debug

# debug.setLogger(debug.Debug('reader', 'compiler'))

inputMibs = ["IF-MIB", "IP-MIB"]
srcDirectories = ["/usr/share/snmp/mibs"]
httpSources = [("mibs.pysnmp.com", 443, "/asn1/@mib@")]


def printOut(mibName, jsonDoc, cbCtx):
    print(f"{os.linesep}{os.linesep}# MIB module {mibName}")
    print(jsonDoc)


# Initialize compiler infrastructure

mibCompiler = MibCompiler(SmiStarParser(), JsonCodeGen(), CallbackWriter(printOut))

# search for source MIBs here
mibCompiler.add_sources(*[FileReader(x) for x in srcDirectories])

# search for source MIBs at Web sites
mibCompiler.add_sources(*[HttpReader(*x) for x in httpSources])

# never recompile MIBs with MACROs
mibCompiler.add_searchers(StubSearcher(*JsonCodeGen.baseMibs))

# run recursive MIB compilation
results = mibCompiler.compile(*inputMibs)

print(f"{os.linesep}# Results: {'', ''.join(f'{x}:{results[x]}' for x in results)}")



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import sys

from pysmi import debug, error
from pysmi.writer.base import AbstractWriter


class CallbackWriter(AbstractWriter):
    """Invokes user-specified callable and passes transformed MIB module to it.

    Note: user callable object signature must be as follows

    .. function:: cbFun(mibname, contents, cbCtx)
    """

    def __init__(self, cbFun, cbCtx=None):
        """Creates an instance of *CallbackWriter* class.

        Args:
            cbFun (callable): user-supplied callable
        Keyword Args:
            cbCtx: user-supplied object passed intact to user callback
        """
        self._cbFun = cbFun
        self._cbCtx = cbCtx

    def __str__(self):
        """Return a string representation of the instance."""
        return f'{self.__class__.__name__}{{"{self._cbFun}"}}'

    def put_data(self, mibname, data, comments=(), dryRun=False):
        if dryRun:
            debug.logger & debug.FLAG_WRITER and debug.logger("dry run mode")
            return

        try:
            self._cbFun(mibname, data, self._cbCtx)

        except Exception:
            raise error.PySmiWriterError(
                f"user callback {self._cbFun} failure writing {mibname}: {sys.exc_info()[1]}",
                writer=self,
            )

        debug.logger & debug.FLAG_WRITER and debug.logger(
            f"user callback for {mibname} succeeded"
        )

    def get_data(self, filename):
        return ""



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import os
import sys
from collections import OrderedDict


import jinja2
from pysmi import debug, error
from pysmi.codegen import jfilters
from pysmi.codegen.intermediate import IntermediateCodeGen

try:
    import json
except ImportError:
    import simplejson as json


class JsonCodeGen(IntermediateCodeGen):
    """Turns MIB AST into JSON document.

    Builds JSON document representing MIB module supplied in form of
    an Abstract Syntax Tree on input.

    Instance of this class is supposed to be passed to *MibCompiler*,
    the rest is internal to *MibCompiler*.
    """

    TEMPLATE_NAME = "jsondoc/base.j2"

    def gen_code(self, ast, symbolTable, **kwargs):
        mibInfo, context = IntermediateCodeGen.gen_code(
            self, ast, symbolTable, **kwargs
        )

        # TODO: reduce code duplication with the other codegens

        searchPath = [os.path.join(os.path.dirname(__file__), "templates")]

        dstTemplate = kwargs.get("dstTemplate")
        if dstTemplate:
            searchPath.insert(0, os.path.dirname(os.path.abspath(dstTemplate)))

        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(searchPath),
            trim_blocks=True,
            lstrip_blocks=True,
        )

        env.filters["capfirst"] = jfilters.capfirst

        try:
            tmpl = env.get_template(dstTemplate or self.TEMPLATE_NAME)
            text = tmpl.render(mib=context)

        except jinja2.exceptions.TemplateError:
            err = sys.exc_info()[1]
            raise error.PySmiCodegenError(f"Jinja template rendering error: {err}")

        debug.logger & debug.FLAG_CODEGEN and debug.logger(
            f"canonical MIB name {mibInfo.name} ({mibInfo.identity}), imported MIB(s) {','.join(mibInfo.imported) or '<none>'}, rendered from {dstTemplate}, JSON document size {len(text)} bytes"
        )

        return mibInfo, text

    # TODO: move this to a template
    def gen_index(self, processed, **kwargs):
        outDict = {
            "meta": {},
            "identity": {},
            "enterprise": {},
            "compliance": {},
            "oids": {},
        }
        if kwargs.get("old_index_data"):
            try:
                outDict.update(json.loads(kwargs["old_index_data"]))

            except Exception:
                raise error.PySmiCodegenError(f"Index load error: {sys.exc_info()[1]}")

        def order(top):
            if isinstance(top, dict):
                new_top = OrderedDict()
                try:
                    # first try to sort keys as OIDs
                    for k in sorted(top, key=lambda x: [int(y) for y in x.split(".")]):
                        new_top[k] = order(top[k])

                except ValueError:
                    for k in sorted(top):
                        new_top[k] = order(top[k])

                return new_top
            elif isinstance(top, list):
                new_top = []
                for e in sorted(set(top)):
                    new_top.append(order(e))

                return new_top

            return top

        for module, status in processed.items():
            modData = outDict["identity"]
            identity_oid = getattr(status, "identity", None)
            if identity_oid:
                if identity_oid not in modData:
                    modData[identity_oid] = []

                modData[identity_oid].append(module)

            modData = outDict["enterprise"]
            enterprise_oid = getattr(status, "enterprise", None)
            if enterprise_oid:
                if enterprise_oid not in modData:
                    modData[enterprise_oid] = []

                modData[enterprise_oid].append(module)

            modData = outDict["compliance"]
            compliance_oids = getattr(status, "compliance", ())
            for compliance_oid in compliance_oids:
                if compliance_oid not in modData:
                    modData[compliance_oid] = []
                modData[compliance_oid].append(module)

            modData = outDict["oids"]
            objects_oids = getattr(status, "oids", ())
            for object_oid in objects_oids:
                if object_oid not in modData:
                    modData[object_oid] = []

                modData[object_oid].append(module)

            if modData:
                unique_prefixes = {}
                for oid in sorted(modData, key=lambda x: x.count(".")):
                    for oid_prefix, modules in unique_prefixes.items():
                        if oid.startswith(oid_prefix) and set(modules).issuperset(
                            modData[oid]
                        ):
                            break
                    else:
                        unique_prefixes[oid] = modData[oid]

                outDict["oids"] = unique_prefixes

        if "comments" in kwargs:
            outDict["meta"]["comments"] = kwargs["comments"]

        debug.logger & debug.FLAG_CODEGEN and debug.logger(
            f"OID->MIB index built, {len(processed)} entries"
        )

        return json.dumps(order(outDict), indent=2)



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import re
import sys
from collections import OrderedDict
from time import strftime, strptime

from pysmi import config, debug, error
from pysmi.codegen.base import AbstractCodeGen
from pysmi.mibinfo import MibInfo


class IntermediateCodeGen(AbstractCodeGen):
    """Turns MIB AST into an intermediate representation.

    This intermediate representation is based on built-in Python types
    and structures that could easily be used from within the template
    engines.
    """

    constImports = {
        "SNMPv2-SMI": (
            "iso",
            "NOTIFICATION-TYPE",  # bug in some MIBs (e.g. A3COM-HUAWEI-DHCPSNOOP-MIB)
            "MODULE-IDENTITY",
            "OBJECT-TYPE",
            "OBJECT-IDENTITY",
        ),
        "SNMPv2-TC": (
            "DisplayString",
            "PhysAddress",
            "TEXTUAL-CONVENTION",
        ),  # XXX
        "SNMPv2-CONF": (
            "MODULE-COMPLIANCE",
            "NOTIFICATION-GROUP",
        ),  # XXX
    }

    # never compile these, they either:
    # - define MACROs (implementation supplies them)
    # - or carry conflicting OIDs (so that all IMPORT's of them will be rewritten)
    # - or have manual fixes
    # - or import base ASN.1 types from implementation-specific MIBs
    fakeMibs = (
        "ASN1",
        "ASN1-ENUMERATION",
        "ASN1-REFINEMENT",
    ) + AbstractCodeGen.baseMibs

    baseTypes = ["Integer", "Integer32", "Bits", "ObjectIdentifier", "OctetString"]

    SMI_TYPES = {
        "NetworkAddress": "IpAddress",  # RFC1065-SMI, RFC1155-SMI -> SNMPv2-SMI
        "nullSpecific": "zeroDotZero",  # RFC1158-MIB -> SNMPv2-SMI
        "ipRoutingTable": "ipRouteTable",  # RFC1158-MIB -> RFC1213-MIB
        "snmpEnableAuthTraps": "snmpEnableAuthenTraps",  # RFC1158-MIB -> SNMPv2-MIB
    }

    def __init__(self):
        self._rows = set()
        self._seenSyms = set()
        self._importMap = {}
        self._out = {}  # k, v = name, generated code
        self._moduleIdentityOid = None
        self._moduleRevision = None
        self._enterpriseOid = None
        self._oids = set()
        self._complianceOids = []
        self.moduleName = ["DUMMY"]
        self.genRules = {"text": True}
        self.symbolTable = {}

    def prep_data(self, pdata):
        data = []
        for el in pdata:
            if not isinstance(el, tuple):
                data.append(el)
            elif len(el) == 1:
                data.append(el[0])
            else:
                data.append(self.handlersTable[el[0]](self, self.prep_data(el[1:])))
        return data

    def gen_imports(self, imports):
        # convertion to SNMPv2
        toDel = []
        for module in list(imports):
            if module in self.convertImportv2:
                for symbol in imports[module]:
                    if symbol in self.convertImportv2[module]:
                        toDel.append((module, symbol))

                        for newImport in self.convertImportv2[module][symbol]:
                            newModule, newSymbol = newImport

                            if newModule in imports:
                                imports[newModule].append(newSymbol)
                            else:
                                imports[newModule] = [newSymbol]

        # removing converted symbols
        for d in toDel:
            imports[d[0]].remove(d[1])

        # merging mib and constant imports
        for module in self.constImports:
            if module in imports:
                imports[module] += self.constImports[module]
            else:
                imports[module] = self.constImports[module]

        outDict = OrderedDict()
        outDict["class"] = "imports"
        for module in sorted(imports):
            symbols = []
            for symbol in sorted(set(imports[module])):
                symbols.append(symbol)

            if symbols:
                self._seenSyms.update([self.trans_opers(s) for s in symbols])
                self._importMap.update([(self.trans_opers(s), module) for s in symbols])
                if module not in outDict:
                    outDict[module] = []

                outDict[module].extend(symbols)

        return OrderedDict(imports=outDict), tuple(sorted(imports))

    def add_to_exports(self, symbol, moduleIdentity=0):
        self._seenSyms.add(symbol)

    # noinspection PyUnusedLocal
    def reg_sym(
        self,
        symbol,
        outDict,
        parentOid=None,
        moduleIdentity=False,
        moduleCompliance=False,
    ):
        if symbol in self._seenSyms and symbol not in self._importMap:
            raise error.PySmiSemanticError(f"Duplicate symbol found: {symbol}")

        self.add_to_exports(symbol, moduleIdentity)
        self._out[symbol] = outDict

        if "oid" in outDict:
            self._oids.add(outDict["oid"])

            if not self._enterpriseOid and outDict["oid"].startswith("1.3.6.1.4.1."):
                self._enterpriseOid = ".".join(outDict["oid"].split(".")[:7])

            if moduleIdentity:
                if self._moduleIdentityOid:
                    if config.STRICT_MODE:
                        raise error.PySmiSemanticError("Duplicate module identity")
                    else:
                        pass
                else:
                    self._moduleIdentityOid = outDict["oid"]

            if moduleCompliance:
                self._complianceOids.append(outDict["oid"])

    def gen_numeric_oid(self, oid):
        numericOid = ()

        for part in oid:
            if isinstance(part, tuple):
                parent, module = part
                if parent == "iso":
                    numericOid += (1,)
                    continue

                if module not in self.symbolTable:
                    # TODO: do getname for possible future borrowed mibs
                    raise error.PySmiSemanticError(
                        f'no module "{module}" in symbolTable'
                    )

                if parent not in self.symbolTable[module]:
                    raise error.PySmiSemanticError(
                        f'no symbol "{parent}" in module "{module}"'
                    )
                numericOid += self.gen_numeric_oid(
                    self.symbolTable[module][parent]["oid"]
                )

            else:
                numericOid += (part,)

        return numericOid

    def get_base_type(self, symName, module):
        if module not in self.symbolTable:
            raise error.PySmiSemanticError(f'no module "{module}" in symbolTable')

        if symName not in self.symbolTable[module]:
            raise error.PySmiSemanticError(
                f'no symbol "{symName}" in module "{module}"'
            )

        symType, symSubtype = self.symbolTable[module][symName].get(
            "syntax", (("", ""), "")
        )
        if not symType[0]:
            raise error.PySmiSemanticError(f'unknown type for symbol "{symName}"')

        if symType[0] in self.baseTypes:
            return symType, symSubtype

        else:
            baseSymType, baseSymSubtype = self.get_base_type(*symType)

            if isinstance(baseSymSubtype, dict):
                # An enumeration (INTEGER or BITS). Combine the enumeration
                # lists when applicable. That is a bit more permissive than
                # strictly needed, as syntax refinement may only remove entries
                # from the base enumeration (RFC 2578 Sec. 9 point (2)).
                if isinstance(symSubtype, dict):
                    baseSymSubtype.update(symSubtype)
                symSubtype = baseSymSubtype

            elif isinstance(baseSymSubtype, list):
                # A value range or size constraint. Note that each list is an
                # intersection of unions of ranges. Taking the intersection
                # instead of the most-top level union of ranges is a bit more
                # restrictive than strictly needed, as range syntax refinement
                # may only remove allowed values from the base type (RFC 2578
                # Sec. 9. points (1) and (3)), but it matches what pyasn1 does.
                if isinstance(symSubtype, list):
                    symSubtype += baseSymSubtype
                else:
                    symSubtype = baseSymSubtype

            return baseSymType, symSubtype

    def is_type_derived_from_tc(self, symName, module):
        """Is the given type derived from a Textual-Convention declaration?

        Given that deriving simple types from textual conventions is currently
        not supported altogether, this method tests only the immediate parent.

        Any problems are ignored: if the given type is not definitely derived
        from a textual convention, this method returns False.
        """
        if module not in self.symbolTable or symName not in self.symbolTable[module]:
            return False

        return self.symbolTable[module][symName]["isTC"]

    # Clause generation functions

    # noinspection PyUnusedLocal
    def gen_agent_capabilities(self, data):
        name, productRelease, status, description, reference, oid = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid

        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = oidStr
        outDict["class"] = "agentcapabilities"

        if productRelease:
            outDict["productrelease"] = productRelease

        if status:
            outDict["status"] = status

        if self.genRules["text"] and description:
            outDict["description"] = description

        if self.genRules["text"] and reference:
            outDict["reference"] = reference

        self.reg_sym(pysmiName, outDict, parentOid)

        return outDict

    # noinspection PyUnusedLocal
    def gen_module_identity(self, data):
        name, lastUpdated, organization, contactInfo, description, revisions, oid = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid

        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = oidStr
        outDict["class"] = "moduleidentity"

        if revisions:
            outDict["revisions"] = revisions

            self._moduleRevision = revisions[0]["revision"]

        if self.genRules["text"]:
            if lastUpdated:
                outDict["lastupdated"] = lastUpdated
            if organization:
                outDict["organization"] = organization
            if contactInfo:
                outDict["contactinfo"] = contactInfo
            if description:
                outDict["description"] = description

        self.reg_sym(pysmiName, outDict, parentOid, moduleIdentity=True)

        return outDict

    # noinspection PyUnusedLocal
    def gen_module_compliance(self, data):
        name, status, description, reference, compliances, oid = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid

        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = oidStr
        outDict["class"] = "modulecompliance"

        if compliances:
            outDict["modulecompliance"] = compliances

        if status:
            outDict["status"] = status

        if self.genRules["text"] and description:
            outDict["description"] = description

        if self.genRules["text"] and reference:
            outDict["reference"] = reference

        self.reg_sym(pysmiName, outDict, parentOid, moduleCompliance=True)

        return outDict

    # noinspection PyUnusedLocal
    def gen_notification_group(self, data):
        name, objects, status, description, reference, oid = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid
        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = oidStr
        outDict["class"] = "notificationgroup"

        if objects:
            outDict["objects"] = [
                {
                    "module": self._importMap.get(
                        self.trans_opers(obj), self.moduleName[0]
                    ),
                    "object": obj,
                }
                for obj in objects
            ]

        if status:
            outDict["status"] = status

        if self.genRules["text"] and description:
            outDict["description"] = description

        if self.genRules["text"] and reference:
            outDict["reference"] = reference

        self.reg_sym(pysmiName, outDict, parentOid)

        return outDict

    # noinspection PyUnusedLocal
    def gen_notification_type(self, data):
        name, objects, status, description, reference, oid = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid
        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = oidStr
        outDict["class"] = "notificationtype"

        if objects:
            outDict["objects"] = [
                {
                    "module": self._importMap.get(
                        self.trans_opers(obj), self.moduleName[0]
                    ),
                    "object": obj,
                }
                for obj in objects
            ]

        if status:
            outDict["status"] = status

        if self.genRules["text"] and description:
            outDict["description"] = description

        if self.genRules["text"] and reference:
            outDict["reference"] = reference

        self.reg_sym(pysmiName, outDict, parentOid)

        return outDict

    # noinspection PyUnusedLocal
    def gen_object_group(self, data):
        name, objects, status, description, reference, oid = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid
        outDict = OrderedDict({"name": name, "oid": oidStr, "class": "objectgroup"})

        if objects:
            outDict["objects"] = [
                {
                    "module": self._importMap.get(
                        self.trans_opers(obj), self.moduleName[0]
                    ),
                    "object": obj,
                }
                for obj in objects
            ]

        if status:
            outDict["status"] = status

        if self.genRules["text"] and description:
            outDict["description"] = description

        if self.genRules["text"] and reference:
            outDict["reference"] = reference

        self.reg_sym(pysmiName, outDict, parentOid)

        return outDict

    # noinspection PyUnusedLocal
    def gen_object_identity(self, data):
        name, status, description, reference, oid = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid

        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = oidStr
        outDict["class"] = "objectidentity"

        if status:
            outDict["status"] = status

        if self.genRules["text"] and description:
            outDict["description"] = description

        if self.genRules["text"] and reference:
            outDict["reference"] = reference

        self.reg_sym(pysmiName, outDict, parentOid)

        return outDict

    # noinspection PyUnusedLocal
    def gen_object_type(self, data):
        (
            name,
            syntax,
            units,
            maxaccess,
            status,
            description,
            reference,
            augmention,
            index,
            defval,
            oid,
        ) = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid
        indexStr, fakeSyms, fakeSymDicts = index or ("", [], [])

        defval = self.process_defval(defval, objname=pysmiName)

        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = oidStr

        if syntax[0]:
            nodetype = syntax[0] == "Bits" and "scalar" or syntax[0]  # Bits hack
            # If this object type is used as a column, but it also has a
            # "SEQUENCE OF" syntax, then it is really a table and not a column.
            #
            # Note that _symtable_cols contains original (non-Pythonized)
            # symbol names! Thus, check with 'name' rather than 'pysmiName'.
            isColumn = (
                name in self.symbolTable[self.moduleName[0]]["_symtable_cols"]
                and syntax[1]
            )
            nodetype = isColumn and "column" or nodetype
            outDict["nodetype"] = nodetype

        outDict["class"] = "objecttype"

        if syntax[1]:
            outDict["syntax"] = syntax[1]
        if defval:
            outDict["default"] = defval
        if units:
            outDict["units"] = units
        if maxaccess:
            outDict["maxaccess"] = maxaccess
        if indexStr:
            outDict["indices"] = indexStr
        if self.genRules["text"] and reference:
            outDict["reference"] = reference
        if augmention:
            augmention = self.trans_opers(augmention)
            outDict["augmention"] = OrderedDict()
            outDict["augmention"]["name"] = name
            outDict["augmention"]["module"] = self.moduleName[0]
            outDict["augmention"]["object"] = augmention
        if status:
            outDict["status"] = status

        if self.genRules["text"] and description:
            outDict["description"] = description

        self.reg_sym(pysmiName, outDict, parentOid)

        for fakeSym, fakeSymDict in zip(fakeSyms, fakeSymDicts):
            fakeSymDict["oid"] = f"{oidStr}.{fakeSymDict['oid']}"
            self.reg_sym(fakeSym, fakeSymDict, pysmiName)

        return outDict

    # noinspection PyUnusedLocal
    def gen_trap_type(self, data):
        name, enterprise, variables, description, reference, value = data

        pysmiName = self.trans_opers(name)

        enterpriseStr, parentOid = enterprise

        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = enterpriseStr + ".0." + str(value)
        outDict["class"] = "notificationtype"

        if variables:
            outDict["objects"] = [
                {
                    "module": self._importMap.get(
                        self.trans_opers(obj), self.moduleName[0]
                    ),
                    "object": obj,
                }
                for obj in variables
            ]

        if self.genRules["text"] and description:
            outDict["description"] = description

        if self.genRules["text"] and reference:
            outDict["reference"] = reference

        self.reg_sym(pysmiName, outDict, parentOid)

        return outDict

    # noinspection PyUnusedLocal
    def gen_type_declaration(self, data):
        name, declaration = data

        outDict = OrderedDict()
        outDict["name"] = name
        outDict["class"] = "type"

        if declaration:
            parentType, attrs = declaration
            if parentType:  # skipping SEQUENCE case
                pysmiName = self.trans_opers(name)
                outDict.update(attrs)
                self.reg_sym(pysmiName, outDict)

                # Establish if this type is derived from a Textual-Convention
                # declaration, as needed for pysnmp code generation.
                typeType = outDict["type"]["type"]
                if (
                    typeType in self.symbolTable[self.moduleName[0]]
                    or typeType in self._importMap
                ):
                    module = self._importMap.get(typeType, self.moduleName[0])
                    isDerivedFromTC = self.is_type_derived_from_tc(typeType, module)
                else:
                    isDerivedFromTC = False
                outDict["type"]["tcbase"] = isDerivedFromTC

        return outDict

    # noinspection PyUnusedLocal
    def gen_value_declaration(self, data):
        name, oid = data

        pysmiName = self.trans_opers(name)

        oidStr, parentOid = oid
        outDict = OrderedDict()
        outDict["name"] = name
        outDict["oid"] = oidStr
        outDict["class"] = "objectidentity"

        self.reg_sym(pysmiName, outDict, parentOid)

        return outDict

    # Subparts generation functions

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_bit_names(self, data):
        names = data[0]
        return names

    def gen_bits(self, data):
        bits = data[0]

        outDict = OrderedDict()
        outDict["type"] = "Bits"
        outDict["class"] = "type"
        outDict["bits"] = OrderedDict()

        for name, bit in sorted(bits, key=lambda x: x[1]):
            outDict["bits"][name] = bit

        return "scalar", outDict

    # noinspection PyUnusedLocal
    def gen_compliances(self, data):
        compliances = []

        for complianceModule in data[0]:
            name = complianceModule[0] or self.moduleName[0]
            compliances += [
                {"object": compl, "module": name} for compl in complianceModule[1]
            ]

        return compliances

    # noinspection PyUnusedLocal
    def gen_conceptual_table(self, data):
        row = data[0]

        if row[1] and row[1][-2:] == "()":
            row = row[1][:-2]
            self._rows.add(row)

        return "table", ""

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_contact_info(self, data):
        text = data[0]
        return self.textFilter("contact-info", text)

    def is_in_range(self, intersection, value):
        """Check whether the given value falls within the given constraints.

        The given list represents an intersection and contains elements that
        represent unions of individual ranges. Each individual range consists
        of one or two elements (a single value, or a low..high range). The
        given value is considered in range if it falls within range for each of
        the unions within the intersection. If the intersection is empty, the
        value is accepted as well.
        """
        for union in intersection:
            in_range = False
            for rng in union:
                if self.str2int(rng[0]) <= value <= self.str2int(rng[-1]):
                    in_range = True
                    break
            if not in_range:
                return False
        return True

    # noinspection PyUnusedLocal
    def gen_display_hint(self, data):
        return data[0]

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_def_val(self, data):
        return data[0]

    def _process_integer_defval(self, defval, defvalType):
        """Process a DEFVAL value for an integer/enumeration type."""
        if isinstance(defval, int):  # decimal
            value = defval

        elif self.is_hex(defval):  # hex
            value = int(defval[1:-2] or "0", 16)

        elif self.is_binary(defval):  # binary
            value = int(defval[1:-2] or "0", 2)

        elif isinstance(defvalType[1], dict):  # enumeration label
            # For enumerations, the ASN.1 DEFVAL statements contain names,
            # whereas the code generation template expects integer values
            # (represented as strings).
            nameToValueMap = defvalType[1]

            # buggy MIB: DEFVAL { { ... } }
            if isinstance(defval, list):
                defval = [dv for dv in defval if dv in nameToValueMap]
                if defval:
                    defval = defval[0]
                # (fall through)

            # good MIB: DEFVAL { ... }
            if defval not in nameToValueMap:
                raise ValueError("unknown enumeration label")

            # Return early so as to skip the enumeration value check below.
            # After all, we already know the resulting number is valid.
            return str(nameToValueMap[defval]), "decimal"

        else:
            raise ValueError("wrong input type for integer")

        if isinstance(defvalType[1], dict):  # enumeration number
            # For numerical values given for enumerated integers, make sure
            # that they are valid, because pyasn1 will not check this case and
            # thus let us set a default value that is not valid for the type.
            if value not in defvalType[1].values():
                raise ValueError("wrong enumeration value")

        elif isinstance(defvalType[1], list):  # range constraints
            if not self.is_in_range(defvalType[1], value):
                raise ValueError("value does not conform to range constraints")

        return str(value), "decimal"

    def _process_string_defval(self, defval, defvalType):
        """Process a DEFVAL value for a string type (including BITS)."""
        defvalBaseType = defvalType[0][0]  # either "OctetString" or "Bits"

        if isinstance(defval, int):  # decimal
            if defvalBaseType != "Bits":
                raise ValueError("decimal values have no meaning for OCTET STRING")

            # Convert the value to a hex string. Add padding if needed.
            value = defval and hex(defval)[2:] or ""
            value = ("0" + value) if len(value) % 2 == 1 else value

            fmt = "hex"

        elif self.is_hex(defval):  # hex
            # Extract the value as hex string. Add padding if needed.
            value = defval[1:-2]
            value = ("0" + value) if len(value) % 2 == 1 else value

            fmt = "hex"

        elif self.is_binary(defval):  # binary
            binval = defval[1:-2]

            # Make sure not to lose any leading zeroes. Add padding if needed.
            width = ((len(binval) + 7) // 8) * 2
            value = width and "{:0{width}x}".format(int(binval, 2), width=width) or ""
            fmt = "hex"

        elif defval and defval[0] == '"' and defval[-1] == '"':  # quoted string
            if defvalBaseType != "OctetString":
                raise ValueError("quoted strings have no meaning for BITS")

            value = defval[1:-1]
            fmt = "string"

        elif defvalBaseType == "Bits" and isinstance(defval, list):  # bit labels
            defvalBits = []

            bits = defvalType[1]

            for bit in defval:
                bitValue = bits.get(bit, None)
                if bitValue is not None:
                    defvalBits.append((bit, bitValue))
                else:
                    raise ValueError("unknown bit")

            return self.gen_bits([defvalBits])[1], "bits"

        else:
            raise ValueError("wrong input type for string")

        if defvalBaseType == "OctetString" and isinstance(
            defvalType[1], list
        ):  # size constraints
            size = len(value) // 2 if fmt == "hex" else len(value)

            if not self.is_in_range(defvalType[1], size):
                raise ValueError("value does not conform to size constraints")

        return value, fmt

    def _process_oid_defval(self, defval, defvalType):
        """Process a DEFVAL value for an object identifier."""
        if isinstance(defval, str) and (
            self.trans_opers(defval) in self.symbolTable[self.moduleName[0]]
            or self.trans_opers(defval) in self._importMap
        ):
            pysmiDefval = self.trans_opers(defval)
            module = self._importMap.get(pysmiDefval, self.moduleName[0])

            value = self.gen_numeric_oid(self.symbolTable[module][pysmiDefval]["oid"])

            return str(value), "oid"

        else:
            raise ValueError("wrong input type for object identifier")

    def process_defval(self, defval, objname):
        if defval is None:
            return {}

        defvalType = self.get_base_type(objname, self.moduleName[0])
        defvalBaseType = defvalType[0][0]

        # Our general policy is that DEFVAL values are considered discardable:
        # any values that were accepted by the parser but turn out to be
        # invalid, are dropped here, so that they will not keep the MIB from
        # being compiled or loaded. The underlying idea is that DEFVAL values
        # are not all that important, and therefore better discarded than the
        # cause of a MIB loading failure.
        #
        # For each combination of input value and object type, the following
        # table shows whether the combination is:
        # - required to be supported by RFC 2578;
        # - accepted out of lenience by our implementation; or,
        # - discarded for being no meaningful combination.
        # Note that enumerations are also Integer32/Integer types, but handled
        # slightly differently, so they are a separate column in this table.
        #
        # Input         Integer     Enumeration OctetString ObjectId.   Bits
        # ------------- ----------- ----------- ----------- ----------- ---------
        # decimal       required    accepted    discarded   discarded   accepted
        # hex           accepted    accepted    required    discarded   accepted
        # binary        accepted    accepted    required    discarded   accepted
        # quoted string discarded   discarded   required    discarded   discarded
        # symbol/label  discarded   required    discarded   required    discarded
        # brackets      discarded   accepted    discarded   discarded   required

        try:
            if defvalBaseType in ("Integer32", "Integer"):
                value, fmt = self._process_integer_defval(defval, defvalType)
            elif defvalBaseType in ("OctetString", "Bits"):
                value, fmt = self._process_string_defval(defval, defvalType)
            elif defvalBaseType == "ObjectIdentifier":
                value, fmt = self._process_oid_defval(defval, defvalType)
            else:  # pragma: no cover
                raise ValueError("unknown base type")

        except ValueError:
            # Discard the given default value.
            return {}

        outDict = OrderedDict(basetype=defvalBaseType, value=value, format=fmt)
        return {"default": outDict}

    # noinspection PyMethodMayBeStatic
    def gen_description(self, data):
        return self.textFilter("description", data[0])

    # noinspection PyMethodMayBeStatic
    def gen_reference(self, data):
        return self.textFilter("reference", data[0])

    # noinspection PyMethodMayBeStatic
    def gen_status(self, data):
        return data[0]

    def gen_product_release(self, data):
        return data[0]

    def gen_enum_spec(self, data):
        items = data[0]
        return {"enumeration": dict(items)}

    # noinspection PyUnusedLocal
    def gen_table_index(self, data):
        def gen_fake_sym_dict(fakeSym, fakeOidSuffix, idxType):
            syntaxDict = OrderedDict()
            syntaxDict["type"] = self.SMI_TYPES.get(idxType, idxType)
            syntaxDict["class"] = "type"

            outDict = OrderedDict()
            outDict["name"] = fakeSym
            outDict["oid"] = str(fakeOidSuffix)  # suffix only; fixed up later
            outDict["class"] = "objecttype"
            outDict["nodetype"] = "column"
            outDict["syntax"] = syntaxDict
            outDict["maxaccess"] = "not-accessible"
            outDict["status"] = "mandatory"  # SMIv1
            return outDict

        indexes = data[0]
        idxStrlist, fakeSyms, fakeSymDicts = [], [], []

        # For fake indices, we generate fake column object types. Each of those
        # is given its own OID. Minimize the chance of collisions by starting
        # with the highest possible OID child number and going backward.
        fakeOidSuffix = 2**32 - 1

        for idx in indexes:
            isImplied = idx[0]
            idxName = idx[1]
            if idxName in self.smiv1IdxTypes:  # SMIv1 support
                idxType = idxName
                idxName = self.fakeIdxPrefix + str(self.fakeIdxNumber)
                fakeSymDict = gen_fake_sym_dict(idxName, fakeOidSuffix, idxType)
                fakeSyms.append(idxName)
                fakeSymDicts.append(fakeSymDict)
                self.fakeIdxNumber += 1
                fakeOidSuffix -= 1

            index = OrderedDict()
            index["module"] = self._importMap.get(
                self.trans_opers(idxName), self.moduleName[0]
            )
            index["object"] = idxName
            index["implied"] = isImplied
            idxStrlist.append(index)

        return idxStrlist, fakeSyms, fakeSymDicts

    def gen_integer_subtype(self, data):
        ranges = []
        for rng in data[0]:
            vmin, vmax = len(rng) == 1 and (rng[0], rng[0]) or rng
            vmin, vmax = self.str2int(vmin), self.str2int(vmax)
            ran = OrderedDict()
            ran["min"] = vmin
            ran["max"] = vmax
            ranges.append(ran)

        return {"range": ranges}

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_max_access(self, data):
        return data[0]

    def gen_octetstring_subtype(self, data):
        sizes = []
        for rng in data[0]:
            vmin, vmax = len(rng) == 1 and (rng[0], rng[0]) or rng
            vmin, vmax = self.str2int(vmin), self.str2int(vmax)

            size = OrderedDict()
            size["min"] = vmin
            size["max"] = vmax
            sizes.append(size)

        outDict = {"size": sizes}

        # If the union of ranges consists of a single size, then store that
        # information as well, so that pysnmp (which needs to know) does not
        # have to compute it itself. We take a slightly elaborate approach so
        # that strange ranges such as SIZE(4 | 4) also produce a fixed size.
        #
        # Note that as per RFC 2578 Sec. 9 point (3), derived types may only
        # ever reduce the ranges of their base types, so we never need to
        # override a base type's fixed length to be "non-fixed" again.
        minSize = min(size["min"] for size in sizes)
        maxSize = max(size["max"] for size in sizes)
        if minSize == maxSize:
            outDict["fixed"] = minSize

        return outDict

    # noinspection PyUnusedLocal
    def gen_oid(self, data):
        out = ()
        parent = ""
        for el in data[0]:
            if isinstance(el, str):
                parent = self.trans_opers(el)
                out += ((parent, self._importMap.get(parent, self.moduleName[0])),)

            elif isinstance(el, int):
                out += (el,)

            elif isinstance(el, tuple):
                out += (el[1],)  # XXX Do we need to create a new object el[0]?

            else:
                raise error.PySmiSemanticError(f"unknown datatype for OID: {el}")

        return ".".join([str(x) for x in self.gen_numeric_oid(out)]), parent

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_objects(self, data):
        return data[0]

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_time(self, data):
        times = []
        for timeStr in data:
            if len(timeStr) == 11:
                timeStr = "19" + timeStr

            elif config.STRICT_MODE and len(timeStr) != 13:
                raise error.PySmiSemanticError(f"Invalid date {timeStr}")
            try:
                times.append(
                    strftime("%Y-%m-%d %H:%M", strptime(timeStr, "%Y%m%d%H%MZ"))
                )

            except ValueError:
                if config.STRICT_MODE:
                    raise error.PySmiSemanticError(f"Invalid date {timeStr}")

                timeStr = "197001010000Z"  # dummy date for dates with typos
                times.append(
                    strftime("%Y-%m-%d %H:%M", strptime(timeStr, "%Y%m%d%H%MZ"))
                )

        return times

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_last_updated(self, data):
        return data[0]

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_organization(self, data):
        return self.textFilter("organization", data[0])

    # noinspection PyUnusedLocal
    def gen_revisions(self, data):
        revisions = []
        for x in data[0]:
            revision = OrderedDict()
            revision["revision"] = self.gen_time([x[0]])[0]
            revision["description"] = self.textFilter("description", x[1][1])
            revisions.append(revision)
        return revisions

    def gen_row(self, data):
        row = data[0]
        row = self.trans_opers(row)

        return (
            row in self.symbolTable[self.moduleName[0]]["_symtable_rows"]
            and ("row", "")
            or self.gen_simple_syntax(data)
        )

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_sequence(self, data):
        return "", ""

    def gen_simple_syntax(self, data):
        objType = data[0]
        objType = self.SMI_TYPES.get(objType, objType)
        objType = self.trans_opers(objType)

        subtype = len(data) == 2 and data[1] or {}

        outDict = OrderedDict()
        outDict["type"] = objType
        outDict["class"] = "type"

        if subtype:
            outDict["constraints"] = subtype

        return "scalar", outDict

    # noinspection PyUnusedLocal
    def gen_type_declaration_rhs(self, data):
        if len(data) == 1:
            parentType, attrs = data[0]

            outDict = OrderedDict()
            if not attrs:
                return outDict
            # just syntax
            outDict["type"] = attrs

        else:
            # Textual convention
            display, status, description, reference, syntax = data
            parentType, attrs = syntax

            outDict = OrderedDict()
            outDict["type"] = attrs
            outDict["class"] = "textualconvention"
            if display:
                outDict["displayhint"] = display
            if status:
                outDict["status"] = status
            if self.genRules["text"] and description:
                outDict["description"] = description
            if self.genRules["text"] and reference:
                outDict["reference"] = reference

        return parentType, outDict

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gen_units(self, data):
        text = data[0]
        return self.textFilter("units", text)

    handlersTable = {
        "agentCapabilitiesClause": gen_agent_capabilities,
        "moduleIdentityClause": gen_module_identity,
        "moduleComplianceClause": gen_module_compliance,
        "notificationGroupClause": gen_notification_group,
        "notificationTypeClause": gen_notification_type,
        "objectGroupClause": gen_object_group,
        "objectIdentityClause": gen_object_identity,
        "objectTypeClause": gen_object_type,
        "trapTypeClause": gen_trap_type,
        "typeDeclaration": gen_type_declaration,
        "valueDeclaration": gen_value_declaration,
        "PRODUCT-RELEASE": gen_product_release,
        "ApplicationSyntax": gen_simple_syntax,
        "BitNames": gen_bit_names,
        "BITS": gen_bits,
        "ComplianceModules": gen_compliances,
        "conceptualTable": gen_conceptual_table,
        "CONTACT-INFO": gen_contact_info,
        "DISPLAY-HINT": gen_display_hint,
        "DEFVAL": gen_def_val,
        "DESCRIPTION": gen_description,
        "REFERENCE": gen_reference,
        "Status": gen_status,
        "enumSpec": gen_enum_spec,
        "INDEX": gen_table_index,
        "integerSubType": gen_integer_subtype,
        "MaxAccessPart": gen_max_access,
        "Notifications": gen_objects,
        "octetStringSubType": gen_octetstring_subtype,
        "objectIdentifier": gen_oid,
        "Objects": gen_objects,
        "LAST-UPDATED": gen_last_updated,
        "ORGANIZATION": gen_organization,
        "Revisions": gen_revisions,
        "row": gen_row,
        "SEQUENCE": gen_sequence,
        "SimpleSyntax": gen_simple_syntax,
        "typeDeclarationRHS": gen_type_declaration_rhs,
        "UNITS": gen_units,
        "VarTypes": gen_objects,
        # 'a': lambda x: genXXX(x, 'CONSTRAINT')
    }

    # TODO: make intermediate format less tied to JSON
    # One thing is to produce OIDs in a tuple form
    # The other thing is index data - may be we should
    # have it prepared at the intermediate stage...?

    def gen_code(self, ast, symbolTable, **kwargs):
        self.genRules["text"] = kwargs.get("genTexts", False)
        self.textFilter = kwargs.get("textFilter") or (
            lambda symbol, text: re.sub(r"\s+", " ", text)
        )
        self.symbolTable = symbolTable
        self._rows.clear()
        self._seenSyms.clear()
        self._importMap.clear()
        self._out.clear()
        self._moduleIdentityOid = None
        self._enterpriseOid = None
        self._oids = set()
        self._complianceOids = []
        self.moduleName[0], moduleOid, imports, declarations = ast

        outDict, importedModules = self.gen_imports(imports)

        for declr in declarations or []:
            if declr:
                self.handlersTable[declr[0]](self, self.prep_data(declr[1:]))

        for sym in self.symbolTable[self.moduleName[0]]["_symtable_order"]:
            if sym not in self._out:
                raise error.PySmiCodegenError(f"No generated code for symbol {sym}")

            outDict[sym] = self._out[sym]

        outDict["meta"] = OrderedDict()
        outDict["meta"]["module"] = self.moduleName[0]

        if "comments" in kwargs:
            outDict["meta"]["comments"] = kwargs["comments"]

        debug.logger & debug.FLAG_CODEGEN and debug.logger(
            f"canonical MIB name {self.moduleName[0]} ({moduleOid}), imported MIB(s) {','.join(importedModules) or '<none>'}"
        )

        return (
            MibInfo(
                oid=moduleOid,
                identity=self._moduleIdentityOid,
                name=self.moduleName[0],
                revision=self._moduleRevision,
                oids=self._oids,
                enterprise=self._enterpriseOid,
                compliance=self._complianceOids,
                imported=tuple(x for x in importedModules if x not in self.fakeMibs),
            ),
            outDict,
        )



#!/usr/bin/env python
#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
# SNMP SMI/MIB data management tool
#
import getopt
import os
import sys
from pathlib import Path

from pysmi import config, debug, error
from pysmi.borrower import AnyFileBorrower, PyFileBorrower
from pysmi.codegen import JsonCodeGen, NullCodeGen, PySnmpCodeGen
from pysmi.compiler import MibCompiler
from pysmi.parser import SmiV1CompatParser
from pysmi.reader import get_readers_from_urls
from pysmi.searcher import (
    AnyFileSearcher,
    PyFileSearcher,
    PyPackageSearcher,
    StubSearcher,
)
from pysmi.writer import CallbackWriter, FileWriter, PyFileWriter


def start():
    # sysexits.h
    EX_OK = 0
    EX_USAGE = 64
    EX_SOFTWARE = 70
    EX_MIB_MISSING = 79
    EX_MIB_FAILED = 79

    # Defaults
    verboseFlag = True
    mibSources = []
    doFuzzyMatchingFlag = True
    mibSearchers = []
    mibStubs = []
    mibBorrowers = []
    dstFormat = None
    dstTemplate = None
    dstDirectory = None
    cacheDirectory = ""
    nodepsFlag = False
    rebuildFlag = False
    dryrunFlag = False
    genMibTextsFlag = False
    keepTextsLayout = False
    pyCompileFlag = True
    pyOptimizationLevel = 0
    ignoreErrorsFlag = False
    buildIndexFlag = False
    writeMibsFlag = True

    helpMessage = f"""\
    Usage: {sys.argv[0]} [--help]
        [--version]
        [--quiet]
        [--strict]
        [--debug=<{"|".join(sorted(debug.FLAG_MAP))}>]
        [--mib-source=<URI>]
        [--mib-searcher=<PATH|PACKAGE>]
        [--mib-stub=<MIB-NAME>]
        [--mib-borrower=<PATH>]
        [--destination-format=<FORMAT>]
        [--destination-template=<PATH>]
        [--destination-directory=<DIRECTORY>]
        [--cache-directory=<DIRECTORY>]
        [--disable-fuzzy-source]
        [--no-dependencies]
        [--no-python-compile]
        [--python-optimization-level]
        [--ignore-errors]
        [--build-index]
        [--rebuild]
        [--dry-run]
        [--no-mib-writes]
        [--generate-mib-texts]
        [--keep-texts-layout]
        <MIB-NAME> [MIB-NAME [...]]]
    Where:
        URI      - file, zip, http, https schemes are supported.
                Use @mib@ placeholder token in URI to refer directly to
                the required MIB module when source does not support
                directory listing (e.g. HTTP).
        FORMAT   - pysnmp, json, null
        TEMPLATE - path to a Jinja2 template extending the base one (see
                documentation for details)"""

    try:
        opts, inputMibs = getopt.getopt(
            sys.argv[1:],
            "hv",
            [
                "help",
                "version",
                "quiet",
                "strict",
                "debug=",
                "mib-source=",
                "mib-searcher=",
                "mib-stub=",
                "mib-borrower=",
                "destination-format=",
                "destination-template=",
                "destination-directory=",
                "cache-directory=",
                "no-dependencies",
                "no-python-compile",
                "python-optimization-level=",
                "ignore-errors",
                "build-index",
                "rebuild",
                "dry-run",
                "no-mib-writes",
                "generate-mib-texts",
                "disable-fuzzy-source",
                "keep-texts-layout",
            ],
        )

    except getopt.GetoptError:
        if verboseFlag:
            sys.stderr.write(
                f"ERROR: {sys.exc_info()[1]}{os.linesep}{helpMessage}{os.linesep}"
            )

        sys.exit(EX_USAGE)

    for opt in opts:
        if opt[0] == "-h" or opt[0] == "--help":
            sys.stderr.write(
                f"""\
    Synopsis:
    SNMP SMI/MIB files conversion tool
    Documentation:
    https://www.pysnmp.com/pysmi
    {helpMessage}
    """
            )
            sys.exit(EX_OK)

        if opt[0] == "-v" or opt[0] == "--version":
            from pysmi import __version__

            sys.stderr.write(
                f"""\
    SNMP SMI/MIB library version {__version__}, written by Ilya Etingof <etingof@gmail.com>
    Python interpreter: {sys.version}
    Software documentation and support at https://www.pysnmp.com/pysmi
    {helpMessage}
    """
            )
            sys.exit(EX_OK)

        if opt[0] == "--quiet":
            verboseFlag = False

        if opt[0] == "--strict":
            config.STRICT_MODE = True

        if opt[0] == "--debug":
            debug.set_logger(debug.Debug(*opt[1].split(",")))

        if opt[0] == "--mib-source":
            mibSources.append(opt[1])

        if opt[0] == "--mib-searcher":
            mibSearchers.append(opt[1])

        if opt[0] == "--mib-stub":
            mibStubs.append(opt[1])

        if opt[0] == "--mib-borrower":
            mibBorrowers.append((opt[1], genMibTextsFlag))

        if opt[0] == "--destination-format":
            dstFormat = opt[1]

        if opt[0] == "--destination-template":
            dstTemplate = opt[1]

        if opt[0] == "--destination-directory":
            dstDirectory = opt[1]

        if opt[0] == "--cache-directory":
            cacheDirectory = opt[1]

        if opt[0] == "--no-dependencies":
            nodepsFlag = True

        if opt[0] == "--no-python-compile":
            pyCompileFlag = False

        if opt[0] == "--python-optimization-level":
            try:
                pyOptimizationLevel = int(opt[1])

            except ValueError:
                sys.stderr.write(
                    f"ERROR: known Python optimization levels: -1, 0, 1, 2{os.linesep}{helpMessage}{os.linesep}"
                )
                sys.exit(EX_USAGE)

        if opt[0] == "--ignore-errors":
            ignoreErrorsFlag = True

        if opt[0] == "--build-index":
            buildIndexFlag = True

        if opt[0] == "--rebuild":
            rebuildFlag = True

        if opt[0] == "--dry-run":
            dryrunFlag = True

        if opt[0] == "--no-mib-writes":
            writeMibsFlag = False

        if opt[0] == "--generate-mib-texts":
            genMibTextsFlag = True

        if opt[0] == "--disable-fuzzy-source":
            doFuzzyMatchingFlag = False

        if opt[0] == "--keep-texts-layout":
            keepTextsLayout = True

    if not mibSources:
        mibSources = [
            "file:///usr/share/snmp/mibs",
            "https://mibs.pysnmp.com/asn1/@mib@",
        ]

    if inputMibs:
        mibSources = (
            sorted(
                {
                    Path(x.replace("file:///", "")).resolve().parent.as_uri()
                    for x in inputMibs
                    if "\\" in x or "/" in x
                }
            )
            + mibSources
        )

        inputMibs = [os.path.basename(os.path.splitext(x)[0]) for x in inputMibs]

    if not inputMibs:
        sys.stderr.write(
            f"ERROR: MIB module names not specified{os.linesep}{helpMessage}{os.linesep}"
        )
        sys.exit(EX_USAGE)

    if not dstFormat:
        dstFormat = "pysnmp"

    if dstFormat == "pysnmp":
        if not mibSearchers:
            mibSearchers = PySnmpCodeGen.defaultMibPackages

        if not mibStubs:
            mibStubs = [
                x for x in PySnmpCodeGen.baseMibs if x not in PySnmpCodeGen.fakeMibs
            ]

        if not mibBorrowers:
            mibBorrowers = [
                ("https://mibs.pysnmp.com:443/mibs/notexts/@mib@", False),
                ("https://mibs.pysnmp.com:443/mibs/fulltexts/@mib@", True),
            ]

        if not dstDirectory:
            dstDirectory = os.path.expanduser("~")
            if sys.platform[:3] == "win":
                dstDirectory = os.path.join(
                    dstDirectory, "PySNMP Configuration", "mibs"
                )
            else:
                dstDirectory = os.path.join(dstDirectory, ".pysnmp", "mibs")

        # Compiler infrastructure

        borrowers = [
            PyFileBorrower(x[1], genTexts=mibBorrowers[x[0]][1])
            for x in enumerate(
                get_readers_from_urls(
                    *[m[0] for m in mibBorrowers], **dict(lowcaseMatching=False)
                )
            )
        ]

        searchers = [PyFileSearcher(dstDirectory)]

        for mibSearcher in mibSearchers:
            searchers.append(PyPackageSearcher(mibSearcher))

        searchers.append(StubSearcher(*mibStubs))

        codeGenerator = PySnmpCodeGen()

        fileWriter = PyFileWriter(dstDirectory).set_options(
            pyCompile=pyCompileFlag, pyOptimizationLevel=pyOptimizationLevel
        )

    elif dstFormat == "json":
        if not mibStubs:
            mibStubs = JsonCodeGen.baseMibs

        if not mibBorrowers:
            mibBorrowers = [
                ("https://mibs.pysnmp.com/json/notexts/@mib@", False),
                ("https://mibs.pysnmp.com/fulltexts/@mib@", True),
            ]

        if not dstDirectory:
            dstDirectory = os.path.join(".")

        # Compiler infrastructure

        borrowers = [
            AnyFileBorrower(x[1], genTexts=mibBorrowers[x[0]][1]).set_options(
                exts=[".json"]
            )
            for x in enumerate(
                get_readers_from_urls(
                    *[m[0] for m in mibBorrowers], **dict(lowcaseMatching=False)
                )
            )
        ]

        searchers = [
            AnyFileSearcher(dstDirectory).set_options(exts=[".json"]),
            StubSearcher(*mibStubs),
        ]

        codeGenerator = JsonCodeGen()

        fileWriter = FileWriter(dstDirectory).set_options(suffix=".json")

    elif dstFormat == "null":
        if not mibStubs:
            mibStubs = NullCodeGen.baseMibs

        if not mibBorrowers:
            mibBorrowers = [
                ("https://mibs.pysnmp.com/null/notexts/@mib@", False),
                ("https://mibs.pysnmp.com/null/fulltexts/@mib@", True),
            ]

        if not dstDirectory:
            dstDirectory = ""

        # Compiler infrastructure

        codeGenerator = NullCodeGen()

        searchers = [StubSearcher(*mibStubs)]

        borrowers = [
            AnyFileBorrower(x[1], genTexts=mibBorrowers[x[0]][1])
            for x in enumerate(
                get_readers_from_urls(
                    *[m[0] for m in mibBorrowers], **dict(lowcaseMatching=False)
                )
            )
        ]

        fileWriter = CallbackWriter(lambda *x: None)

    else:
        sys.stderr.write(
            f"ERROR: unknown destination format: {dstFormat}{os.linesep}{helpMessage}{os.linesep}"
        )
        sys.exit(EX_USAGE)

    if verboseFlag:
        sys.stderr.write(
            f"""\
Source MIB repositories: {', '.join(mibSources)}
Borrow missing/failed MIBs from: {', '.join([x[0] for x in mibBorrowers if x[1] == genMibTextsFlag])}
Existing/compiled MIB locations: {', '.join(mibSearchers)}
Compiled MIBs destination directory: {dstDirectory}
MIBs excluded from code generation: {', '.join(sorted(mibStubs))}
MIBs to compile: {', '.join(inputMibs)}
Destination format: {dstFormat}
Custom destination template: {dstTemplate}
Parser grammar cache directory: {cacheDirectory or "not used"}
Also compile all relevant MIBs: {"no" if nodepsFlag else "yes"}
Rebuild MIBs regardless of age: {"yes" if rebuildFlag else "no"}
Dry run mode: {"yes" if dryrunFlag else "no"}
Create/update MIBs: {"yes" if writeMibsFlag else "no"}
Byte-compile Python modules: {"yes" if dstFormat == "pysnmp" and pyCompileFlag else "no"} (optimization level {"yes" if dstFormat == "pysnmp" and pyOptimizationLevel else "no"})
Ignore compilation errors: {"yes" if ignoreErrorsFlag else "no"}
Generate OID->MIB index: {"yes" if buildIndexFlag else "no"}
Generate texts in MIBs: {"yes" if genMibTextsFlag else "no"}
Keep original texts layout: {"yes" if keepTextsLayout else "no"}
Try various file names while searching for MIB module: {"yes" if doFuzzyMatchingFlag else "no"}
"""
        )

    # Initialize compiler infrastructure

    mibCompiler = MibCompiler(
        SmiV1CompatParser(tempdir=cacheDirectory), codeGenerator, fileWriter
    )

    try:
        mibCompiler.add_sources(
            *get_readers_from_urls(
                *mibSources, **dict(fuzzyMatching=doFuzzyMatchingFlag)
            )
        )

        mibCompiler.add_searchers(*searchers)

        mibCompiler.add_borrowers(*borrowers)

        processed = mibCompiler.compile(
            *inputMibs,
            **dict(
                noDeps=nodepsFlag,
                rebuild=rebuildFlag,
                dryRun=dryrunFlag,
                dstTemplate=dstTemplate,
                genTexts=genMibTextsFlag,
                textFilter=keepTextsLayout and (lambda symbol, text: text) or None,
                writeMibs=writeMibsFlag,
                ignoreErrors=ignoreErrorsFlag,
            ),
        )

        safe = {}
        sorted_files = sorted(processed)
        for x in sorted_files:
            if processed[x] != "failed":
                safe[x] = processed[x]

        if buildIndexFlag:
            mibCompiler.build_index(safe, dryRun=dryrunFlag, ignoreErrors=True)

    except error.PySmiError:
        sys.stderr.write(f"ERROR: {sys.exc_info()[1]}{os.linesep}")
        sys.exit(EX_SOFTWARE)

    else:
        compiled = [x for x in sorted_files if processed[x] == "compiled"]
        borrowed = [x for x in sorted_files if processed[x] == "borrowed"]
        untouched = [x for x in sorted_files if processed[x] == "untouched"]
        missing = [x for x in sorted_files if processed[x] == "missing"]
        unprocessed = [x for x in sorted_files if processed[x] == "unprocessed"]
        failed = [x for x in sorted_files if processed[x] == "failed"]
        if verboseFlag:
            sys.stdout.write(
                "{}reated/updated MIBs: {}{}".format(
                    dryrunFlag and "Would be c" or "C",
                    ", ".join(
                        [
                            f"{x} ({processed[x].alias})"
                            if x != processed[x].alias
                            else f"{x}"
                            for x in compiled
                        ]
                    ),
                    os.linesep,
                )
            )

            sys.stdout.write(
                "Pre-compiled MIBs {}borrowed: {}{}".format(
                    dryrunFlag and "would be " or "",
                    ", ".join([f"{x} ({processed[x].path})" for x in borrowed]),
                    os.linesep,
                )
            )

            sys.stdout.write(
                "Up to date MIBs: {}{}".format(
                    ", ".join([f"{x}" for x in untouched]),
                    os.linesep,
                )
            )
            sys.stderr.write(
                "Missing source MIBs: {}{}".format(
                    f"{os.linesep} ".join([f"{x}" for x in missing]),
                    os.linesep,
                )
            )

            sys.stderr.write(
                "Ignored MIBs: {}{}".format(
                    ", ".join([f"{x}" for x in unprocessed]),
                    os.linesep,
                )
            )

            sys.stderr.write(
                "Failed MIBs: {}{}".format(
                    f"{os.linesep} ".join(
                        [f"{x} ({processed[x].error})" for x in failed]
                    ),
                    os.linesep,
                )
            )

        exitCode = EX_OK

        if len(missing) > 0:
            exitCode = EX_MIB_MISSING

        if len(failed) > 0:
            exitCode = EX_MIB_FAILED

        sys.exit(exitCode)



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import sys
import time

from pysmi import __version__ as pysmi_version
from pysmi import debug
from pysmi import error
from pysmi.compat import decode
from pysmi.mibinfo import MibInfo
from pysmi.reader.base import AbstractReader
from requests import session


class HttpReader(AbstractReader):
    """Fetch ASN.1 MIB text by name from a web site.

    *HttpReader* class instance tries to download ASN.1 MIB files
    by name and return their contents to caller.
    """

    MIB_MAGIC = "@mib@"

    def __init__(self, url):
        """Create an instance of *HttpReader* bound to specific URL.

        Note:
            The `http_proxy` and `https_proxy` environment variables are
            respected by the underlying `urllib` stdlib module.

        Args:
            host (str): domain name or IP address of web server
            port (int): TCP port web server is listening
            locationTemplate (str): location part of the URL optionally containing @mib@ magic placeholder to be replaced with MIB name. If @mib@ magic is not present, MIB name is appended to `locationTemplate`

        Keyword Args:
            timeout (int): response timeout
            ssl (bool): access HTTPS web site
        """
        self._url = url

        self.session = session()

        self._user_agent = f"pysmi-{pysmi_version}; python-{sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}; {sys.platform}"

    def __str__(self):
        """Return string representation of the instance."""
        return self._url

    def get_data(self, mibname, **options):
        headers = {"Accept": "text/plain", "User-Agent": self._user_agent}

        mibname = decode(mibname)

        debug.logger & debug.FLAG_READER and debug.logger(f"looking for MIB {mibname}")

        for mibalias, mibfile in self.get_mib_variants(mibname, **options):
            if self.MIB_MAGIC in self._url:
                url = self._url.replace(self.MIB_MAGIC, mibfile)
            else:
                url = self._url + mibfile

            debug.logger & debug.FLAG_READER and debug.logger(
                f"trying to fetch MIB from {url}"
            )

            try:
                response = self.session.get(url, headers=headers)

            except Exception:
                debug.logger & debug.FLAG_READER and debug.logger(
                    f"failed to fetch MIB from {url}: {sys.exc_info()[1]}"
                )
                continue

            debug.logger & debug.FLAG_READER and debug.logger(
                f"HTTP response {response.status_code}"
            )

            if response.status_code == 200:
                try:
                    mtime = time.mktime(
                        time.strptime(
                            response.headers["Last-Modified"],
                            "%a, %d %b %Y %H:%M:%S %Z",
                        )
                    )

                except Exception:
                    debug.logger & debug.FLAG_READER and debug.logger(
                        f"malformed HTTP headers: {sys.exc_info()[1]}"
                    )
                    mtime = time.time()

                debug.logger & debug.FLAG_READER and debug.logger(
                    f"fetching source MIB {url}, mtime {response.headers['Last-Modified']}"
                )

                return MibInfo(
                    path=url, file=mibfile, name=mibalias, mtime=mtime
                ), response.content.decode("utf-8")

        raise error.PySmiReaderFileNotFoundError(
            f"source MIB {mibname} not found", reader=self
        )



"""
Compile MIBs from web
+++++++++++++++++++++

Look up specific ASN.1 MIBs at configured Web sites,
compile them into pysnmp form and save Python modules as plain-text
files in a local directory.

Try to support both SMIv1 and SMIv2 flavors of SMI as well as
popular deviations from official syntax found in the wild.

In this example we disable automatic dependency checking on MIB
compilation using noDeps flag.

Also, we do not check if target file already exists thus MIB
compilation occurs on every invocation.
"""  #
from pysmi.reader import HttpReader
from pysmi.searcher import StubSearcher
from pysmi.writer import PyFileWriter
from pysmi.parser import SmiStarParser
from pysmi.codegen import PySnmpCodeGen
from pysmi.compiler import MibCompiler

inputMibs = ["IF-MIB", "IP-MIB"]
httpSources = [("mibs.pysnmp.com", 443, "/asn1/@mib@")]
dstDirectory = ".pysnmp-mibs"

# Initialize compiler infrastructure

mibCompiler = MibCompiler(SmiStarParser(), PySnmpCodeGen(), PyFileWriter(dstDirectory))

# search for source MIBs at Web sites
mibCompiler.add_sources(HttpReader("https://mibs.pysnmp.com/asn1/@mib@"))

# never recompile MIBs with MACROs
mibCompiler.add_searchers(StubSearcher(*PySnmpCodeGen.baseMibs))

# run non-recursive MIB compilation
results = mibCompiler.compile(*inputMibs, **dict(noDeps=True))

print(f"Results: {', '.join(f'{x}:{results[x]}' for x in results)}")



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import re

import ply.lex as lex
from pysmi import config, error
from pysmi import debug
from pysmi.lexer.base import AbstractLexer

UNSIGNED32_MAX = 4294967295
UNSIGNED64_MAX = 18446744073709551615
LEX_VERSION = [int(x) for x in lex.__version__.split(".")]


# Do not overload single lexer methods - overload all or none of them!
# noinspection PySingleQuotedDocstring,PyMethodMayBeStatic,PyIncorrectDocstring
class SmiV2Lexer(AbstractLexer):
    reserved_words = [
        "ACCESS",
        "AGENT-CAPABILITIES",
        "APPLICATION",
        "AUGMENTS",
        "BEGIN",
        "BITS",
        "CONTACT-INFO",
        "CREATION-REQUIRES",
        "Counter",
        "Counter32",
        "Counter64",
        "DEFINITIONS",
        "DEFVAL",
        "DESCRIPTION",
        "DISPLAY-HINT",
        "END",
        "ENTERPRISE",
        "EXTENDS",
        "FROM",
        "GROUP",
        "Gauge",
        "Gauge32",
        "IDENTIFIER",
        "IMPLICIT",
        "IMPLIED",
        "IMPORTS",
        "INCLUDES",
        "INDEX",
        "INSTALL-ERRORS",
        "INTEGER",
        "Integer32",
        "IpAddress",
        "LAST-UPDATED",
        "MANDATORY-GROUPS",
        "MAX-ACCESS",
        "MIN-ACCESS",
        "MODULE",
        "MODULE-COMPLIANCE",
        "MODULE-IDENTITY",
        "NOTIFICATION-GROUP",
        "NOTIFICATION-TYPE",
        "NOTIFICATIONS",
        "OBJECT",
        "OBJECT-GROUP",
        "OBJECT-IDENTITY",
        "OBJECT-TYPE",
        "OBJECTS",
        "OCTET",
        "OF",
        "ORGANIZATION",
        "Opaque",
        "PIB-ACCESS",
        "PIB-DEFINITIONS",
        "PIB-INDEX",
        "PIB-MIN-ACCESS",
        "PIB-REFERENCES",
        "PIB-TAG",
        "POLICY-ACCESS",
        "PRODUCT-RELEASE",
        "REFERENCE",
        "REVISION",
        "SEQUENCE",
        "SIZE",
        "STATUS",
        "STRING",
        "SUBJECT-CATEGORIES",
        "SUPPORTS",
        "SYNTAX",
        "TEXTUAL-CONVENTION",
        "TimeTicks",
        "TRAP-TYPE",
        "UNIQUENESS",
        "UNITS",
        "UNIVERSAL",
        "Unsigned32",
        "VALUE",
        "VARIABLES",
        "VARIATION",
        "WRITE-SYNTAX",
    ]

    reserved = {}
    for w in reserved_words:
        reserved[w] = w.replace("-", "_").upper()
        # hack to support SMIv1
        if w == "Counter":
            reserved[w] = "COUNTER32"
        elif w == "Gauge":
            reserved[w] = "GAUGE32"

    forbidden_words = [
        "ABSENT",
        "ANY",
        "BIT",
        "BOOLEAN",
        "BY",
        "COMPONENT",
        "COMPONENTS",
        "DEFAULT",
        "DEFINED",
        "ENUMERATED",
        "EXPLICIT",
        "EXTERNAL",
        "FALSE",
        "MAX",
        "MIN",
        "MINUS-INFINITY",
        "NULL",
        "OPTIONAL",
        "PLUS-INFINITY",
        "PRESENT",
        "PRIVATE",
        "REAL",
        "SET",
        "TAGS",
        "TRUE",
        "WITH",
    ]

    # Token names required!
    tokens = list(
        set(
            [
                "BIN_STRING",
                "CHOICE",
                "COLON_COLON_EQUAL",
                "DOT_DOT",
                "EXPORTS",
                "HEX_STRING",
                "LOWERCASE_IDENTIFIER",
                "MACRO",
                "NEGATIVENUMBER",
                "NEGATIVENUMBER64",
                "NUMBER",
                "NUMBER64",
                "QUOTED_STRING",
                "UPPERCASE_IDENTIFIER",
            ]
            + list(reserved.values())
        )
    )

    states = (
        ("macro", "exclusive"),
        ("choice", "exclusive"),
        ("exports", "exclusive"),
        ("comment", "exclusive"),
    )

    literals = "[]{}():;,-.|"

    t_DOT_DOT = r"\.\."
    t_COLON_COLON_EQUAL = r"::="

    t_ignore = " \t"

    def __init__(self, tempdir=""):
        self._tempdir = tempdir
        self.lexer = None
        self.reset()

    def reset(self):
        if LEX_VERSION < [3, 0]:
            self.lexer = lex.lex(
                module=self, reflags=re.DOTALL, outputdir=self._tempdir, debug=False
            )
        else:
            if debug.logger & debug.FLAG_LEXER:
                logger = debug.logger.get_current_logger()
            else:
                logger = lex.NullLogger()

            if debug.logger & debug.FLAG_GRAMMAR:
                debuglogger = debug.logger.get_current_logger()
            else:
                debuglogger = None

            self.lexer = lex.lex(
                module=self,
                reflags=re.DOTALL,
                outputdir=self._tempdir,
                debuglog=debuglogger,
                errorlog=logger,
            )

    def t_newline(self, t):
        r"\r\n|\n|\r"
        t.lexer.lineno += 1

    # Skipping MACRO
    def t_MACRO(self, t):
        r"MACRO"
        t.lexer.begin("macro")
        return t

    def t_macro_newline(self, t):
        r"\r\n|\n|\r"
        t.lexer.lineno += 1

    def t_macro_END(self, t):
        r"END"
        t.lexer.begin("INITIAL")
        return t

    def t_macro_body(self, t):
        r".+?(?=END)"
        pass

    # Skipping EXPORTS
    def t_EXPORTS(self, t):
        r"EXPORTS"
        t.lexer.begin("exports")
        return t

    def t_exports_newline(self, t):
        r"\r\n|\n|\r"
        t.lexer.lineno += 1

    def t_exports_end(self, t):
        r";"
        t.lexer.begin("INITIAL")

    def t_exports_body(self, t):
        r"[^;]+"
        pass

    # Skipping CHOICE
    def t_CHOICE(self, t):
        r"CHOICE"
        t.lexer.begin("choice")
        return t

    def t_choice_newline(self, t):
        r"\r\n|\n|\r"
        t.lexer.lineno += 1

    def t_choice_end(self, t):
        r"\}"
        t.lexer.begin("INITIAL")

    def t_choice_body(self, t):
        r"[^\}]+"
        pass

    # Comment handling
    def t_begin_comment(self, t):
        r"--"
        t.lexer.begin("comment")

    def t_comment_newline(self, t):
        r"\r\n|\n|\r"
        t.lexer.lineno += 1
        t.lexer.begin("INITIAL")

    #  def t_comment_end(self, t):
    #    r'--'
    #    t.lexer.begin('INITIAL')

    def t_comment_body(self, t):
        r"[^\r\n]+"
        pass

    def t_UPPERCASE_IDENTIFIER(self, t):
        r"[A-Z][-a-zA-z0-9]*"
        if t.value in self.forbidden_words:
            raise error.PySmiLexerError(f"{t.value} is forbidden", lineno=t.lineno)

        if t.value[-1] == "-":
            raise error.PySmiLexerError(
                f"Identifier should not end with '-': {t.value}", lineno=t.lineno
            )

        t.type = self.reserved.get(t.value, "UPPERCASE_IDENTIFIER")

        return t

    def t_LOWERCASE_IDENTIFIER(self, t):
        r"[0-9]*[a-z][-a-zA-z0-9]*"
        if t.value[-1] == "-":
            raise error.PySmiLexerError(
                f"Identifier should not end with '-': {t.value}", lineno=t.lineno
            )
        return t

    def t_NUMBER(self, t):
        r"-?[0-9]+"
        t.value = int(t.value)
        neg = 0
        if t.value < 0:
            neg = 1

        val = abs(t.value)

        if val <= UNSIGNED32_MAX:
            if neg:
                t.type = "NEGATIVENUMBER"

        elif val <= UNSIGNED64_MAX:
            if neg:
                t.type = "NEGATIVENUMBER64"
            else:
                t.type = "NUMBER64"

        else:
            raise error.PySmiLexerError(f"Number {t.value} is too big", lineno=t.lineno)

        return t

    def t_BIN_STRING(self, t):
        r"'[01]*'[bB]"
        value = t.value[1:-2]
        while value and value[0] == "0" and len(value) % 8:
            value = value[1:]

            if config.STRICT_MODE and len(value) % 8:
                raise error.PySmiLexerError(
                    f"Number of 0s and 1s have to divide by 8 in binary string {t.value}",
                    lineno=t.lineno,
                )

        return t

    def t_HEX_STRING(self, t):
        r"'[0-9a-fA-F]*'[hH]"
        value = t.value[1:-2]
        while value and value[0] == "0" and len(value) % 2:
            value = value[1:]

            if config.STRICT_MODE and len(value) % 2:
                raise error.PySmiLexerError(
                    f"Number of symbols have to be even in hex string {t.value}",
                    lineno=t.lineno,
                )

        return t

    def t_QUOTED_STRING(self, t):
        r"\"[^\"]*\""
        t.lexer.lineno += len(re.findall(r"\r\n|\n|\r", t.value))
        return t

    def t_error(self, t):
        raise error.PySmiLexerError(
            f"Illegal character '{t.value[0]}', {len(t.value) - 1} characters left unparsed at this stage",
            lineno=t.lineno,
        )
        # t.lexer.skip(1)


class SupportSmiV1Keywords:
    @staticmethod
    def reserved():
        reserved_words = [
            "ACCESS",
            "AGENT-CAPABILITIES",
            "APPLICATION",
            "AUGMENTS",
            "BEGIN",
            "BITS",
            "CONTACT-INFO",
            "CREATION-REQUIRES",
            "Counter",
            "Counter32",
            "Counter64",
            "DEFINITIONS",
            "DEFVAL",
            "DESCRIPTION",
            "DISPLAY-HINT",
            "END",
            "ENTERPRISE",
            "EXTENDS",
            "FROM",
            "GROUP",
            "Gauge",
            "Gauge32",
            "IDENTIFIER",
            "IMPLICIT",
            "IMPLIED",
            "IMPORTS",
            "INCLUDES",
            "INDEX",
            "INSTALL-ERRORS",
            "INTEGER",
            "Integer32",
            "IpAddress",
            "LAST-UPDATED",
            "MANDATORY-GROUPS",
            "MAX-ACCESS",
            "MIN-ACCESS",
            "MODULE",
            "MODULE-COMPLIANCE",
            "MAX",
            "MODULE-IDENTITY",
            "NetworkAddress",
            "NOTIFICATION-GROUP",
            "NOTIFICATION-TYPE",
            "NOTIFICATIONS",
            "OBJECT",
            "OBJECT-GROUP",
            "OBJECT-IDENTITY",
            "OBJECT-TYPE",
            "OBJECTS",
            "OCTET",
            "OF",
            "ORGANIZATION",
            "Opaque",
            "PIB-ACCESS",
            "PIB-DEFINITIONS",
            "PIB-INDEX",
            "PIB-MIN-ACCESS",
            "PIB-REFERENCES",
            "PIB-TAG",
            "POLICY-ACCESS",
            "PRODUCT-RELEASE",
            "REFERENCE",
            "REVISION",
            "SEQUENCE",
            "SIZE",
            "STATUS",
            "STRING",
            "SUBJECT-CATEGORIES",
            "SUPPORTS",
            "SYNTAX",
            "TEXTUAL-CONVENTION",
            "TimeTicks",
            "TRAP-TYPE",
            "UNIQUENESS",
            "UNITS",
            "UNIVERSAL",
            "Unsigned32",
            "VALUE",
            "VARIABLES",
            "VARIATION",
            "WRITE-SYNTAX",
        ]

        reserved = {}
        for w in reserved_words:
            reserved[w] = w.replace("-", "_").upper()
            # hack to support SMIv1
            if w == "Counter":
                reserved[w] = "COUNTER32"
            elif w == "Gauge":
                reserved[w] = "GAUGE32"

        return reserved

    @staticmethod
    def forbidden_words():
        return [
            "ABSENT",
            "ANY",
            "BIT",
            "BOOLEAN",
            "BY",
            "COMPONENT",
            "COMPONENTS",
            "DEFAULT",
            "DEFINED",
            "ENUMERATED",
            "EXPLICIT",
            "EXTERNAL",
            "FALSE",
            "MIN",
            "MINUS-INFINITY",
            "NULL",
            "OPTIONAL",
            "PLUS-INFINITY",
            "PRESENT",
            "PRIVATE",
            "REAL",
            "SET",
            "TAGS",
            "TRUE",
            "WITH",
        ]

    @staticmethod
    def tokens():
        # Token names required!
        tokens = [
            "BIN_STRING",
            "CHOICE",
            "COLON_COLON_EQUAL",
            "DOT_DOT",
            "EXPORTS",
            "HEX_STRING",
            "LOWERCASE_IDENTIFIER",
            "MACRO",
            "NEGATIVENUMBER",
            "NEGATIVENUMBER64",
            "NUMBER",
            "NUMBER64",
            "QUOTED_STRING",
            "UPPERCASE_IDENTIFIER",
        ]
        tokens += list(SupportSmiV1Keywords.reserved().values())
        return list(set(tokens))


relaxedGrammar = {
    "supportSmiV1Keywords": [
        SupportSmiV1Keywords.reserved,
        SupportSmiV1Keywords.forbidden_words,
        SupportSmiV1Keywords.tokens,
    ],
    "supportIndex": [],
    "commaAtTheEndOfImport": [],
    "commaAtTheEndOfSequence": [],
    "mixOfCommasAndSpaces": [],
    "uppercaseIdentifier": [],
    "lowcaseIdentifier": [],
    "curlyBracesAroundEnterpriseInTrap": [],
    "noCells": [],
}


def lexerFactory(**grammarOptions):
    classAttr = {}

    for option in grammarOptions:
        if grammarOptions[option]:
            if option not in relaxedGrammar:
                raise error.PySmiError(f"Unknown lexer relaxation option: {option}")

            for func in relaxedGrammar[option]:
                classAttr[func.__name__] = func()

    return type("SmiLexer", (SmiV2Lexer,), classAttr)



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
from pysmi.searcher.anyfile import AnyFileSearcher
from pysmi.searcher.pyfile import PyFileSearcher
from pysmi.searcher.pypackage import PyPackageSearcher
from pysmi.searcher.stub import StubSearcher



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
from pysmi import debug, error
from pysmi.reader.base import AbstractReader


class AbstractBorrower:
    genTexts = False
    exts = ""
    _reader: "AbstractReader"

    def __init__(self, reader, genTexts=False):
        """Creates an instance of *Borrower* class.

        Args:
            reader: a *reader* object

        Keyword Args:
            genText: indicates whether this borrower should be looking
                     for transformed MIBs that include human-oriented texts
        """
        if genTexts is not None:
            self.genTexts = genTexts

        self._reader = reader

    def __str__(self):
        """Return a string representation of the instance."""
        return f"{self.__class__.__name__}{{{self._reader}, genTexts={self.genTexts}, exts={self.exts}}}"

    def set_options(self, **kwargs):
        self._reader.set_options(**kwargs)

        for k in kwargs:
            setattr(self, k, kwargs[k])

        return self

    def get_data(self, mibname, **options):
        if bool(options.get("genTexts")) != self.genTexts:
            debug.logger & debug.FLAG_BORROWER and debug.logger(
                f"skipping incompatible borrower {self} for file {mibname}"
            )
            raise error.PySmiFileNotFoundError(mibname=mibname, reader=self._reader)

        debug.logger & debug.FLAG_BORROWER and (
            debug.logger(f"trying to borrow file {mibname} from {self._reader}")
        )

        if "exts" not in options:
            options["exts"] = self.exts

        return self._reader.get_data(mibname, **options)



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
from pysmi.parser.dialect import smi_v1_relaxed
from pysmi.parser.smi import parserFactory

# compatibility stub
SmiV1CompatParser = parserFactory(**smi_v1_relaxed)
SmiStarParser = SmiV1CompatParser



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
from pysmi import debug
from pysmi.codegen.base import AbstractCodeGen
from pysmi.mibinfo import MibInfo


class NullCodeGen(AbstractCodeGen):
    """Dummy code generation backend.

    Could be used for disabling code generation at *MibCompiler*.
    """

    def gen_code(self, ast, symbolTable, **kwargs):
        debug.logger & debug.FLAG_CODEGEN and debug.logger(
            f"{self.__class__.__name__} invoked"
        )
        return MibInfo(oid=None, name="", imported=[]), ""

    def gen_index(self, mibsMap, **kwargs):
        return ""



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
from pysmi import debug
from pysmi import error
from pysmi.searcher.base import AbstractSearcher


class StubSearcher(AbstractSearcher):
    """Figures out if given MIB module is present in a fixed list of modules."""

    def __init__(self, *mibnames):
        """Create an instance of *StubSearcher* initialized with a fixed list or MIB modules names.

        Args:
            mibnames (str): blacklisted MIB names
        """
        self._mibnames = mibnames

    def __str__(self):
        """Return a string representation of the instance."""
        return self.__class__.__name__

    def file_exists(self, mibname, mtime, rebuild=False):
        if mibname in self._mibnames:
            debug.logger & debug.FLAG_SEARCHER and debug.logger(
                f"pretend compiled {mibname} exists and is very new"
            )
            raise error.PySmiFileNotModifiedError(
                f"compiled file {mibname} is among {', '.join(self._mibnames)}",
                searcher=self,
            )

        raise error.PySmiFileNotFoundError(
            f"no compiled file {mibname} found among {', '.join(self._mibnames)}",
            searcher=self,
        )



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import importlib.machinery

from pysmi.borrower.base import AbstractBorrower


SOURCE_SUFFIXES = importlib.machinery.EXTENSION_SUFFIXES


class PyFileBorrower(AbstractBorrower):
    """Create PySNMP MIB file borrowing object."""

    exts = SOURCE_SUFFIXES
